"""Utilities for the demessaging module."""

# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

import inspect
import textwrap
from typing import Any, Dict, Type

from pydantic import BaseModel


def type_to_string(type_: Any):
    if inspect.isclass(type_):
        return object_to_string(type_)
    else:
        return str(type_)


def object_to_string(obj: Any):
    if obj.__module__ == "builtins":
        return obj.__name__
    return f"{obj.__module__}.{obj.__name__}"


def build_parameter_docs(model: Type[BaseModel]) -> str:
    """Build the docstring for the parameters of a model."""
    docstring = "\n\nParameters\n----------"
    for fieldname, field_info in model.model_fields.items():
        param_doc = textwrap.dedent(
            f"""
            {fieldname} : {type_to_string(field_info.annotation)}
                {field_info.description}
            """
        )
        docstring = docstring + param_doc.rstrip()
    return docstring


def append_parameter_docs(model: Type[BaseModel]) -> Type[BaseModel]:
    """Append the parameters section to the docstring of a model."""
    docstring = build_parameter_docs(model)
    model.__doc__ += docstring  # type: ignore
    return model


def merge_config(base: Dict, to_merge: Dict) -> Dict:
    """Merge two configuration dictionaries.

    Parameters
    ----------
    base : Dict
        The base dictionary that `to_merge` shall be merged into.
    to_merge : Dict
        The dictionary to merge.

    Returns
    -------
    Dict
        `base` merged with `to_merge`

    Notes
    -----
    `base` is modified in-place!
    """
    for key, val in to_merge.items():
        if key not in base:
            base[key] = val
        elif isinstance(val, dict):
            merge_config(base[key], val)
        elif isinstance(val, str):
            base[key] = val
        else:
            try:
                iter(val)
            except TypeError:
                base[key] = val
            else:
                base[key] = list(base[key]) + list(val)
    return base



"""Configuration classes for DASF."""

# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

from .api import configure, registry  # noqa: F401
from .backend import (  # noqa: F401
    BaseConfig,
    ClassConfig,
    FunctionConfig,
    ModuleConfig,
)
from .logging import LoggingConfig  # noqa: F401
from .messaging import (  # noqa: F401
    BaseMessagingConfig,
    PulsarConfig,
    WebsocketURLConfig,
)
from .registry import ApiRegistry  # noqa: F401



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Data Analytics Software Framework

python module wrapper for the data analytics software framework DASF
"""

from __future__ import annotations

from demessaging.backend import BackendModule, main  # noqa: F401
from demessaging.config import configure, registry

from . import _version

__all__ = ["main", "configure", "registry", "BackendModule"]


__version__ = _version.get_versions()["version"]

__author__ = "Daniel Eggert, Mike Sips, Philipp S. Sommer, Doris Dransch"
__copyright__ = "2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences"
__credits__ = [
    "Daniel Eggert",
    "Mike Sips",
    "Philipp S. Sommer",
    "Doris Dransch",
]
__license__ = "Apache-2.0"

__maintainer__ = "Philipp S. Sommer"
__email__ = "philipp.sommer@hereon.de"

__status__ = "Pre-Alpha"



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Test module for the :mod:`demessaging.backend` module."""
import importlib
import inspect
import json
import pathlib
import subprocess as spr
from textwrap import dedent
from typing import Callable, Dict, List, Type, Union

import pytest

from demessaging import backend
from demessaging.messaging.constants import MessageType, PropertyKeys

try:
    import xarray
except ImportError:
    xarray = None  # type: ignore
else:
    try:
        import scipy
    except ImportError:
        scipy = None  # type: ignore


class TestModuleModel:
    """Test class for the :class:`demessaging.backend.BackendModule`."""

    def test_load_all(self) -> None:
        """Test loading everything with __all__."""
        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )
        schema = Model.model_json_schema()

        assert "FuncFuncBasic" in schema["$defs"]

        assert "ClassClass" in schema["$defs"]
        assert "MethClassClassAdd2a" in schema["$defs"]

        assert "_private_func" not in schema["$defs"]
        assert "_PrivateFunc" not in schema["$defs"]
        assert "PrivateFunc" not in schema["$defs"]

    def test_json_schema_extra(self) -> None:
        """Test adding json_schema_extra."""
        """Test loading everything with __all__."""
        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            json_schema_extra={"testModuleExtra": {"testModule": "attribute"}},
        )
        schema = Model.model_json_schema()
        assert "testModuleExtra" in schema
        assert schema["testModuleExtra"] == {"testModule": "attribute"}

    def test_load_function_name(self) -> None:
        """Test loading everything with __all__."""
        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            members=["func_basic"],
        )
        schema = Model.model_json_schema()

        assert "FuncFuncBasic" in schema["$defs"]
        assert "ClassClass" not in schema["$defs"]

    def test_load_function(self) -> None:
        """Test loading everything with __all__."""
        from _test_module import func_basic

        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            members=[func_basic],
        )
        schema = Model.model_json_schema()

        assert "FuncFuncBasic" in schema["$defs"]
        assert "ClassClass" not in schema["$defs"]

    def test_load_function_model(self) -> None:
        """Test loading everything with __all__."""
        from _test_module import func_basic

        FuncModel = backend.BackendFunction.create_model(func_basic)
        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            members=[FuncModel],
        )
        schema = Model.model_json_schema()

        assert "FuncFuncBasic" in schema["$defs"]
        assert "ClassClass" not in schema["$defs"]

    def test_parse_basic(self) -> None:
        """Test parsing a request to the :func:`func_basic` function."""
        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )

        obj = {"func_name": "func_basic", "a": 1}

        model = Model.model_validate(obj)
        val = model()
        assert val.root == [1]  # type: ignore

    def test_parse_class(self) -> None:
        """Test parsing a request for a class method."""
        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )

        obj = {
            "class_name": "Class",
            "a": 2,
            "function": {"func_name": "add2a", "c": 2},
        }

        model = Model.model_validate(obj)
        val = model()
        assert val.root == 4  # type: ignore

    def test_connect(
        self,
        random_topic: str,
        get_test_module_path: Callable[[str], str],
        test_dasf_connect: Callable[[str, str], str],
    ) -> None:
        """Test connecting to the messaging server."""
        test_dasf_connect(random_topic, get_test_module_path("_test_module"))

    def test_request(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_request_path: Callable[[str], str],
        test_dasf_request: Callable[[str, str, Union[Dict, str]], str],
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        modpath = get_test_module_path("_test_module")
        connect_module(random_topic, modpath)
        test_request = get_request_path("test_request")
        test_dasf_request(random_topic, modpath, test_request)

    def test_request_json(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_module_command: Callable[[str, str], List[str]],
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        modpath = get_test_module_path("_test_module")
        connect_module(random_topic, modpath)
        command = get_module_command(random_topic, modpath)
        spr.check_call(
            command
            + ["send-request", json.dumps({"func_name": "func_basic", "a": 1})]
        )

    def test_request_dump_to(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_request_path: Callable[[str], str],
        test_dasf_request: Callable[[str, str, Union[Dict, str]], str],
        tmpdir,
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        modpath = get_test_module_path("_test_module")
        connect_module(random_topic, modpath, "--dump-to", str(tmpdir))  # type: ignore[call-arg]
        test_request = get_request_path("test_request")
        test_dasf_request(random_topic, modpath, test_request)
        assert tmpdir.listdir()
        assert len(tmpdir.listdir()) == 1
        import json

        with open(test_request) as f:
            ref = json.load(f)
        with tmpdir.listdir()[0].open() as f:
            test = json.load(f)
        assert ref == test

    def test_request_dump_tool(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_request_path: Callable[[str], str],
        test_dasf_request: Callable[[str, str, Union[Dict, str]], str],
        tmpdir,
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        modpath = get_test_module_path("_test_module")
        connect_module(  # type: ignore[call-arg]
            random_topic,
            modpath,
            "--dump-tool",
            "cp {} %s/{basename}" % tmpdir,
        )
        test_request = get_request_path("test_request")
        test_dasf_request(random_topic, modpath, test_request)
        assert tmpdir.listdir()
        assert len(tmpdir.listdir()) == 1
        import json

        with open(test_request) as f:
            ref = json.load(f)
        with tmpdir.listdir()[0].open() as f:
            test = json.load(f)
        assert ref == test

    def test_request_cmd(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_request_path: Callable[[str], str],
        test_dasf_request: Callable[[str, str, Union[Dict, str]], str],
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        modpath = get_test_module_path("_test_module")
        connect_module(random_topic, modpath, "--cmd", 'echo "[500]"')  # type: ignore[call-arg]
        test_request = get_request_path("test_request")
        response = test_dasf_request(random_topic, modpath, test_request)
        assert "[500]" in response.splitlines()

    def test_process_request(
        self,
        get_request_path: Callable[[str], str],
    ) -> None:
        """Test processing a request"""
        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )
        test_request = get_request_path("test_request")
        with open(test_request) as f:
            result = Model.process_request(f).root
        assert result == [1]

    def test_cli_compute(
        self,
        random_topic: str,
        get_test_module_path: Callable[[str], str],
        get_request_path: Callable[[str], str],
        get_module_command: Callable[[str, str], List[str]],
    ) -> None:
        """Test running the compute cli command."""
        modpath = get_test_module_path("_test_module")
        command = get_module_command(random_topic, modpath)  # type: ignore[call-arg]
        test_request = get_request_path("test_request")
        response = spr.check_output(command + ["compute", test_request])
        assert response.decode("utf-8").strip() == "[1]"

    def test_cli_compute_json(
        self,
        random_topic: str,
        get_test_module_path: Callable[[str], str],
        get_module_command: Callable[[str, str], List[str]],
    ) -> None:
        """Test running the compute cli command."""
        modpath = get_test_module_path("_test_module")
        command = get_module_command(random_topic, modpath)  # type: ignore[call-arg]
        response = spr.check_output(
            command
            + ["compute", json.dumps({"func_name": "func_basic", "a": 1})]
        )
        assert response.decode("utf-8").strip() == "[1]"

    def test_request_arbitrary(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_request_path: Callable[[str], str],
        test_dasf_request: Callable[[str, str, Union[Dict, str]], str],
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        modpath = get_test_module_path("_test_module")
        connect_module(random_topic, modpath)
        test_request = get_request_path("test_request_arbitrary")
        test_dasf_request(random_topic, modpath, test_request)

    @pytest.mark.skipif(
        xarray is None or scipy is None,
        reason="xarray and netCDF4 are required",
    )
    def test_request_xarray(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_request_path: Callable[[str], str],
        test_dasf_request: Callable[[str, str, Union[Dict, str]], str],
    ) -> None:
        """Test parsing a request with xarray via the messaging system."""
        modpath = get_test_module_path("_test_module_with_xarray")
        connect_module(random_topic, modpath)
        test_request = get_request_path("test_request_xarray")
        test_dasf_request(random_topic, modpath, test_request)

    def test_request_report(
        self,
        random_topic: str,
        connect_module: Callable[..., spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_request_path: Callable[[str], str],
        test_dasf_request: Callable[..., str],
        modified_env: Callable,
    ) -> None:
        """Test parsing a request via the messaging system with a report."""
        modpath = get_test_module_path("_test_module_with_report")
        connect_module(
            random_topic, modpath, env=modified_env(TEST_TOPIC=random_topic)
        )

        test_request = get_request_path("test_report_request")

        test_dasf_request(
            random_topic,
            modpath,
            test_request,
            env=modified_env(TEST_TOPIC=random_topic),
        )

    def test_render(
        self, default_class, tmp_module: pathlib.Path, random_mod_name: str
    ) -> None:
        """Test the rendering of a module."""
        import _test_module as ref

        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )

        code = Model.backend_config.render()

        tmp_module.write_text(code, "utf-8")

        mod = importlib.import_module(random_mod_name)

        # test the Class member
        assert hasattr(mod, "Class")

        Class: Type[ref.Class] = mod.Class  # type: ignore

        ref_doc = dedent(inspect.getdoc(ref.Class)).strip()  # type: ignore
        func_doc = dedent(inspect.getdoc(Class)).strip()  # type: ignore

        assert func_doc == ref_doc

        init_ref_doc = dedent(inspect.getdoc(ref.Class.__init__)).strip()  # type: ignore  # noqa: E501
        init_doc = dedent(inspect.getdoc(Class.__init__)).strip()  # type: ignore  # noqa: E501

        assert init_ref_doc == init_doc

        assert hasattr(Class, "sum") and callable(Class.sum)

        params_ref = list(inspect.signature(ref.Class.sum).parameters)
        params = list(inspect.signature(Class.sum).parameters)

        assert params == params_ref

        # test the function member
        assert hasattr(mod, "func_basic")

        ref_doc = dedent(inspect.getdoc(ref.func_basic)).strip()  # type: ignore  # noqa: E501
        func_doc = dedent(inspect.getdoc(mod.func_basic)).strip()  # type: ignore  # noqa: E501

        assert func_doc == ref_doc

    def test_render_with_type(
        self,
        tmp_module: pathlib.Path,
        random_mod_name: str,
    ) -> None:
        """Test the rendering of a module with a custom type defined in it."""

        Model = backend.BackendModule.create_model(
            "_test_module_with_type", messaging_config=dict(topic="test")
        )

        code = Model.generate()

        tmp_module.write_text(code, "utf-8")

        mod = importlib.import_module(random_mod_name)

        # test the Class member
        assert hasattr(mod, "MyType")

        ini = mod.MyType(a=1)  # type: ignore
        assert ini.add2a(1) == 2

    def test_generate_and_call_basemodel(
        self,
        get_test_module_path: Callable[[str], str],
        connect_module: Callable[..., spr.Popen],
        get_module_command: Callable[[str, str], List[str]],
        tmp_module: pathlib.Path,
        random_mod_name: str,
        random_topic: str,
        modified_env: Callable,
    ) -> None:
        """Test the generation of a frontend API."""

        modpath = get_test_module_path("_test_module_with_basemodel_type")

        command = get_module_command(random_topic, modpath)

        with tmp_module.open("w") as f:
            spr.check_call(
                command + ["generate"],
                stdout=f,
                env=modified_env(TEST_TOPIC=random_topic),
            )

        connect_module(
            random_topic,
            modpath,
            env=modified_env(TEST_TOPIC=random_topic),
        )

        mod = importlib.import_module(random_mod_name)

        # test the Class member
        assert hasattr(mod, "TestBaseModelType")

        result = mod.func_basic([dict(data=2)])  # type: ignore
        assert result == 2

        result = mod.TestClass([dict(data=2)]).test_func(dict(data=2))  # type: ignore
        assert result == 4

    def test_generate_and_call(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_module_command: Callable[[str, str], List[str]],
        tmp_module: pathlib.Path,
        random_mod_name: str,
    ) -> None:
        """Test the generation of a frontend API."""
        import _test_module as ref

        modpath = get_test_module_path("_test_module")

        command = get_module_command(random_topic, modpath)

        with tmp_module.open("w") as f:
            spr.check_call(command + ["generate"], stdout=f)

        connect_module(random_topic, modpath)

        mod = importlib.import_module(random_mod_name)

        # test the function member
        assert hasattr(mod, "func_basic")

        result = mod.func_basic(1)  # type: ignore
        assert result == ref.func_basic(1)

        # test the Class member
        assert hasattr(mod, "Class")

        Class: Type[ref.Class] = mod.Class  # type: ignore

        ini = Class(1)
        ref_ini = ref.Class(1)
        result = ini.add2a(2)
        assert result == ref_ini.add2a(2)

    def test_generate_and_call_report(
        self,
        random_topic: str,
        connect_module: Callable[..., spr.Popen],
        get_test_module_path: Callable[[str], str],
        get_module_command: Callable[[str, str], List[str]],
        tmp_module: pathlib.Path,
        random_mod_name: str,
        modified_env: Callable,
    ) -> None:
        """Test the generation of a frontend API."""
        import _test_module_with_report as ref

        modpath = get_test_module_path("_test_module_with_report")

        command = get_module_command(random_topic, modpath)

        with tmp_module.open("w") as f:
            spr.check_call(
                command + ["generate"],
                stdout=f,
                env=modified_env(TEST_TOPIC=random_topic),
            )

        connect_module(
            random_topic,
            modpath,
            env=modified_env(TEST_TOPIC=random_topic),
        )

        mod = importlib.import_module(random_mod_name)

        # test the function member
        assert hasattr(mod, "report_test")

        TestReport: Type[ref.TestReport] = mod.TestReport  # type: ignore

        result: TestReport = mod.report_test()  # type: ignore
        assert isinstance(result, TestReport)  # type: ignore

        assert len(TestReport._reports) == 2


class TestModuleAPIModel:
    """Tests for the ``ModuleAPIModel`` class."""

    def test_api_info_load_all(self) -> None:
        """Test loading everything with __all__."""
        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )

        api_info: backend.ModuleAPIModel = Model.get_api_info()

        schema = api_info.rpc_schema

        assert "FuncFuncBasic" in schema["$defs"]

        assert "ClassClass" in schema["$defs"]
        assert "MethClassClassAdd2a" in schema["$defs"]

        assert "_private_func" not in schema["$defs"]
        assert "_PrivateFunc" not in schema["$defs"]
        assert "PrivateFunc" not in schema["$defs"]

        classes = api_info.classes

        assert len(classes) == 1
        assert classes[0].name == "Class"

        functions = api_info.functions

        assert len(functions) == 3
        assert functions[0].name == "func_arbitrary_model"
        assert functions[1].name == "func_basic"
        assert functions[2].name == "func_arbitrary_type"

    def test_api_info_load_function_name(self) -> None:
        """Test loading everything with __all__."""
        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            members=["func_basic"],
        )

        api_info = Model.get_api_info()

        schema = api_info.rpc_schema

        assert "FuncFuncBasic" in schema["$defs"]
        assert "ClassClass" not in schema["$defs"]

        classes = api_info.classes

        assert len(classes) == 0

        functions = api_info.functions

        assert len(functions) == 1
        assert functions[0].name == "func_basic"

    def test_api_info_load_function(self) -> None:
        """Test loading everything with __all__."""
        from _test_module import func_basic

        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            members=[func_basic],
        )

        api_info: backend.ModuleAPIModel = Model.get_api_info()

        schema = api_info.rpc_schema

        assert "FuncFuncBasic" in schema["$defs"]
        assert "ClassClass" not in schema["$defs"]

        classes = api_info.classes

        assert len(classes) == 0

        functions = api_info.functions

        assert len(functions) == 1
        assert functions[0].name == "func_basic"

    def test_api_info_request(
        self,
        random_topic: str,
        connect_module: Callable[[str, str], spr.Popen],
        get_test_module_path: Callable[[str], str],
        live_ws_server,
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        from dasf_broker import app_settings

        modpath = get_test_module_path("_test_module")
        connect_module(random_topic, modpath)

        websocket_url = "%s/%s" % (
            live_ws_server.ws_url,
            app_settings.DASF_WEBSOCKET_URL_ROUTE,
        )

        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(
                topic=random_topic, websocket_url=websocket_url
            ),
            members=["func_basic"],
        )

        producer = Model.backend_config.messaging_config.producer

        response = backend.utils.run_async(
            producer.send_request,
            {"properties": {PropertyKeys.MESSAGE_TYPE: MessageType.API_INFO}},
        )

        api_info = backend.ModuleAPIModel.model_validate_json(response["msg"])

        schema = api_info.rpc_schema

        assert "FuncFuncBasic" in schema["$defs"]

        assert "ClassClass" in schema["$defs"]
        assert "MethClassClassAdd2a" in schema["$defs"]

        assert "_private_func" not in schema["$defs"]
        assert "_PrivateFunc" not in schema["$defs"]
        assert "PrivateFunc" not in schema["$defs"]

        classes = api_info.classes

        assert len(classes) == 1
        assert classes[0].name == "Class"

        functions = api_info.functions

        assert len(functions) == 3
        assert functions[0].name == "func_arbitrary_model"
        assert functions[1].name == "func_basic"
        assert functions[2].name == "func_arbitrary_type"



"""Messaging configuration classes for DASF."""

# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import atexit
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from pydantic import Field, Json, PositiveInt, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from demessaging.utils import append_parameter_docs

if TYPE_CHECKING:
    from demessaging.messaging.producer import MessageProducer


@append_parameter_docs
class BaseMessagingConfig(BaseSettings):  # type: ignore
    """Base class for messaging configs."""

    model_config = SettingsConfigDict(env_prefix="de_backend_")

    topic: str = Field(
        description=(
            "The topic identifier under which to register at the pulsar."
        )
    )

    header: Union[Json[Dict[str, Any]], Dict[str, Any]] = Field(  # type: ignore
        default_factory=dict, description="Header parameters for the request"  # type: ignore[arg-type]
    )

    max_workers: Optional[PositiveInt] = Field(
        default=None,
        description=(
            "(optional) number of concurrent workers for handling requests, "
            "default: number of processors on the machine, multiplied by 5."
        ),
    )

    queue_size: Optional[PositiveInt] = Field(
        default=None,
        description=(
            "(optional) size of the request queue, if MAX_WORKERS is set, "
            "this needs to be at least as big as MAX_WORKERS, "
            "otherwise an AttributeException is raised."
        ),
    )

    max_payload_size: int = Field(
        default=500 * 1024,
        description=(
            "(optional) maximum payload size, must be smaller than pulsars 'webSocketMaxTextFrameSize', "
            "which is configured e.g.via 'pulsar/conf/standalone.conf'."
            "default: 512000 (500kb)."
        ),
    )

    producer_keep_alive: int = Field(
        default=120,
        description=(
            "The amount of time that the websocket connection to a producer "
            "should be kept open. By default, 2 minutes (120 seconds). On "
            "each outgoing message, the timer will be reset. Set this to 0 to "
            "immediately close the connection when a message has been sent "
            "and acknowledged."
        ),
    )

    producer_connection_timeout: int = Field(
        default=30,
        description=(
            "The amount of time that we grant producers to establish a "
            "connection to the message broker in order to send a response. If "
            "a connection cannot be established in this time, the response "
            "will not be sent and the connection will be closed."
        ),
    )

    _producer: Optional[MessageProducer] = None

    @property
    def producer(self) -> MessageProducer:
        """The connected producer for the messaging config"""
        from demessaging.messaging.producer import MessageProducer

        if self._producer:
            return self._producer
        else:
            self._producer = producer = MessageProducer(self)
            producer.connect()
            atexit.register(producer.disconnect)
            return producer

    @model_validator(mode="after")
    def validate_queue_size(self):
        """Check that the queue_size is smaller than the max_workers."""
        queue_size = self.queue_size
        max_workers = self.max_workers
        if queue_size is not None and max_workers is not None:
            if queue_size < max_workers:
                raise ValueError(
                    f"queue_size ({queue_size}) needs to be larger than or "
                    f"equal to max_workers ({max_workers})"
                )
        return self

    def get_topic_url(
        self, topic: str, subscription: Optional[str] = None
    ) -> str:
        """Build the URL to connect to a websocket."""
        raise NotImplementedError(
            "this method is supposed to be implemented by subclasses"
        )


@append_parameter_docs
class PulsarConfig(BaseMessagingConfig):
    """A configuration class to connect to the pulsar messaging framework."""

    host: str = Field(
        "localhost", description="The remote host of the pulsar."
    )

    port: str = Field(
        "8080", description="The port of the pulsar at the given :attr:`host`."
    )

    persistent: str = Field("non-persistent")

    tenant: str = Field("public")

    namespace: str = Field("default")

    def get_topic_url(
        self, topic: str, subscription: Optional[str] = None
    ) -> str:
        """Build the URL to connect to a websocket."""
        connection_type = "consumer" if subscription else "producer"
        sub = ("/" + subscription) if subscription else ""
        return (
            f"ws://{self.host}:{self.port}/ws/v2/{connection_type}/"
            f"{self.persistent}/{self.tenant}/{self.namespace}/{topic}{sub}"
        )


@append_parameter_docs
class WebsocketURLConfig(BaseMessagingConfig):
    """A configuration for a websocket."""

    websocket_url: str = Field(
        "", description="The fully qualified URL to the websocket."
    )

    producer_url: Optional[str] = Field(
        None,
        description=(
            "An alternative URL to use for producers. If None, the "
            "`websocket_url` will be used."
        ),
    )

    consumer_url: Optional[str] = Field(
        None,
        description=(
            "An alternative URL to use for consumers. If None, the "
            "`websocket_url` will be used."
        ),
    )

    def get_topic_url(
        self, topic: str, subscription: Optional[str] = None
    ) -> str:
        """Build the URL to connect to a websocket."""
        sub = ("/" + subscription) if subscription else ""
        if subscription:
            uri = self.consumer_url or self.websocket_url
        else:
            uri = self.producer_url or self.websocket_url
        if not uri.endswith("/"):
            uri += "/"
        return uri + topic + sub



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Validators for xarray datatypes

This module contains validators for the `xarray` datatype.
"""

from base64 import b64decode
from typing import Any

import xarray as xr


def validate_dataarray(cls, v: Any) -> xr.DataArray:
    """Validate that a value can be coerced to an xarray.DataArray

    This function transforms the input `v` into an :class:`xarray.DataArray`.

    Usage in DASF::

        from pydantic import field_validator
        from demessaging import configure
        from demessaging.validators.xarray import validate_dataarray

        import xarray as xr

        @configure(
            validators={
                "da_validator": field_validator("da")(validate_dataarray)
            }
        )
        def some_function(da: xr.DataArray):
            ...


    Parameters
    ----------
    v : Any
        Value that should be transformed into a :class:`~xarray.DataArray`.
        The following types are accepted:

        :class:`dict`
            If `v` is a dictionary, it will be passed to
            :meth:`xarray.DataArray.from_dict`
        :class:`str`
            If `v` is a string, we assume that it is a base64-encoded
            utf-8 decoded string from
            :func:`demessaging.serializers.xarray.encode_xarray`. This will be
            decoded and transformed into a dataarray using the ``scipy``
            or ``h5netcdf`` engine in :func:`xarray.open_dataarray`
        :class:`bytes`
            ``bytes`` are directly passed to :func:`xarray.open_dataarray`
        anything else
            if None of the above mentioned objects are used, we pass `v` to
            the constructor of :class:`xarray.DataArray`

    Returns
    -------
    xarray.DataArray
        The DataArray the has been created from `v`
    """
    if isinstance(v, dict):
        return xr.DataArray.from_dict(v)
    elif isinstance(v, str):
        return xr.open_dataarray(b64decode(v.encode("utf-8")))  # type: ignore
    elif isinstance(v, bytes):
        return xr.open_dataarray(v)  # type: ignore
    else:
        return xr.DataArray(v)


def validate_dataset(cls, v: Any) -> xr.Dataset:
    """Validate that a value can be coerced to an xarray.Dataset

    This function transforms the input `v` into an :class:`xarray.Dataset`.

    Usage in DASF::

        from pydantic import field_validator
        from demessaging import configure
        from demessaging.validators.xarray import validate_dataset

        import xarray as xr

        @configure(
            validators={
                "ds_validator": field_validator("ds")(validate_dataset)
            }
        )
        def some_function(ds: xr.Dataset):
            ...


    Parameters
    ----------
    v : Any
        Value that should be transformed into a :class:`~xarray.Dataset`.
        The following types are accepted:

        :class:`dict`
            If `v` is a dictionary, it will be passed to
            :meth:`xarray.Dataset.from_dict`
        :class:`str`
            If `v` is a string, we assume that it is a base64-encoded
            utf-8 decoded string from
            :func:`demessaging.serializers.xarray.encode_xarray`. This will be
            decoded and transformed into a dataarray using the ``scipy``
            or ``h5netcdf`` engine in :func:`xarray.open_dataset`
        :class:`bytes`
            ``bytes`` are directly passed to :func:`xarray.open_dataset`
        anything else
            if None of the above mentioned objects are used, we pass `v` to
            the constructor of :class:`xarray.Dataset`

    Returns
    -------
    xarray.Dataset
        The Dataset the has been created from `v`
    """
    if isinstance(v, dict):
        return xr.Dataset.from_dict(v)
    elif isinstance(v, str):
        return xr.open_dataset(b64decode(v.encode("utf-8")))  # type: ignore
    elif isinstance(v, bytes):
        return xr.open_dataset(v)  # type: ignore
    else:
        return xr.Dataset(v)



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Test module for :mod:`demessaging.config`."""
from typing import Type

from demessaging import backend, config


def test_configure_function(func_sig) -> None:
    """Test configurating a function."""
    config.configure(field_params={"a": {"gt": 0}})(func_sig)
    Model = backend.BackendFunction.create_model(func_sig)

    schema = Model.model_json_schema()

    assert "exclusiveMinimum" in schema["properties"]["a"]
    assert schema["properties"]["a"]["exclusiveMinimum"] == 0


def test_configure_function_merge(func_sig) -> None:
    """Test merging multiple configurations for a function."""
    config.configure(field_params={"a": {"gt": 0}})(func_sig)
    config.configure(field_params={"a": {"lt": 1}})(func_sig)
    Model = backend.BackendFunction.create_model(func_sig)

    schema = Model.model_json_schema()

    assert "exclusiveMinimum" in schema["properties"]["a"]
    assert schema["properties"]["a"]["exclusiveMinimum"] == 0
    assert schema["properties"]["a"]["exclusiveMaximum"] == 1


def test_configure_class(default_class) -> None:
    """Test configurating a class."""
    config.configure(field_params={"a": {"gt": 0}})(default_class)
    Model = backend.BackendClass.create_model(default_class)

    schema = Model.model_json_schema()

    assert "exclusiveMinimum" in schema["properties"]["a"]
    assert schema["properties"]["a"]["exclusiveMinimum"] == 0


def test_configure_class_merge(default_class) -> None:
    """Test merging multiple configurations for a class."""
    config.configure(field_params={"a": {"gt": 0}})(default_class)
    config.configure(field_params={"a": {"lt": 1}})(default_class)
    Model = backend.BackendClass.create_model(default_class)

    schema = Model.model_json_schema()

    assert "exclusiveMinimum" in schema["properties"]["a"]
    assert schema["properties"]["a"]["exclusiveMinimum"] == 0
    assert schema["properties"]["a"]["exclusiveMaximum"] == 1


def test_configure_method(default_class: Type[object]) -> None:
    """Test configuring a specific method."""

    class TestClass(default_class):  # type: ignore
        @config.configure(field_params={"a": {"gt": 0}})
        def test_method(self, a: int) -> int:
            return a

    Model = backend.BackendClass.create_model(TestClass)

    schema = Model.model_json_schema()

    assert (
        "exclusiveMinimum"
        in schema["$defs"]["MethClassTestClassTestMethod"]["properties"]["a"]
    )
    aconf = schema["$defs"]["MethClassTestClassTestMethod"]["properties"]["a"]
    assert aconf["exclusiveMinimum"] == 0


def test_configure_method_merge(default_class: Type[object]) -> None:
    """Test merging multiple configurations for a specific method."""

    class TestClass(default_class):  # type: ignore
        @config.configure(field_params={"a": {"lt": 1}})
        @config.configure(field_params={"a": {"gt": 0}})
        def test_method(self, a: int) -> int:
            return a

    Model = backend.BackendClass.create_model(TestClass)

    schema = Model.model_json_schema()

    assert (
        "exclusiveMinimum"
        in schema["$defs"]["MethClassTestClassTestMethod"]["properties"]["a"]
    )
    aconf = schema["$defs"]["MethClassTestClassTestMethod"]["properties"]["a"]
    assert aconf["exclusiveMinimum"] == 0
    assert aconf["exclusiveMaximum"] == 1



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Configuration classes for the de-messaging backend module."""
from __future__ import annotations

import inspect
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    Union,
)
from warnings import warn

from deprogressapi import BaseReport
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import ConfigDict, Field, ImportString
from pydantic.functional_serializers import PlainSerializer
from typing_extensions import Annotated

from demessaging.config.logging import LoggingConfig
from demessaging.config.messaging import PulsarConfig, WebsocketURLConfig
from demessaging.config.registry import ApiRegistry
from demessaging.template import Template
from demessaging.utils import append_parameter_docs, object_to_string

if TYPE_CHECKING:
    from demessaging.backend.class_ import BackendClass
    from demessaging.backend.function import BackendFunction


def _get_registry() -> ApiRegistry:
    """Convenience function to get :attr:`registry`.

    Without this, the default value for the config classes `registry` attribute
    would always be empty.
    """
    import demessaging.config

    return demessaging.config.registry


@append_parameter_docs
class BaseConfig(BaseModel):
    """Configuration base class for functions, modules and classes."""

    doc: str = Field(
        "",
        description=(
            "The documentation of the object. If empty, this will be taken "
            "from the corresponding ``__doc__`` attribute."
        ),
    )

    registry: ApiRegistry = Field(
        default_factory=_get_registry,
        description="Utilities for imports and encoders.",
    )

    template: Template = Field(
        Template(name="empty"),  # type: ignore
        description=(
            "The :class:`demessaging.template.Template` that is used "
            "to render this object for the generated API."
        ),
    )

    def render(self, **context) -> str:
        """Generate the code to call this function in the frontend."""
        context["config"] = self
        code = self.template.render(**context)
        return code


@append_parameter_docs
class FunctionConfig(BaseConfig):
    """Configuration class for a backend module function."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    name: str = Field(
        "",
        description=(
            "The name of the function. If empty, this will be taken from the "
            "functions ``__name__`` attribute."
        ),
    )

    signature: Optional[inspect.Signature] = Field(
        None,
        description=(
            "The calling signature for the function. If empty, this will be "
            "taken from the function itself."
        ),
    )

    validators: Dict[
        str,
        List[
            Union[
                ImportString,
                Annotated[Callable, PlainSerializer(object_to_string)],
            ]
        ],
    ] = Field(
        default_factory=dict,
        description=(
            "Custom validators for function arguments. This parameter is a "
            "mapping from function argument name to a list of callables that "
            "can be used as validator."
        ),
    )

    serializers: Dict[
        str,
        Union[
            ImportString,
            Annotated[Callable, PlainSerializer(object_to_string)],
        ],
    ] = Field(
        default_factory=dict,
        description=(
            "A mapping from function argument to serializing function that is "
            "then used for the "
            ":class:`pydantic.functional_serializers.PlainSerializer`."
        ),
    )

    return_validators: Optional[
        List[
            Union[
                ImportString,
                Annotated[Callable, PlainSerializer(object_to_string)],
            ]
        ]
    ] = Field(
        None,
        description=(
            "Validators for the return value. This parameter is a list of "
            "callables that can be used as validator for the return value."
        ),
    )

    return_serializer: Optional[
        Union[
            ImportString,
            Annotated[Callable, PlainSerializer(object_to_string)],
        ]
    ] = Field(
        None,
        description=("A function that is used to serialize the return value."),
    )

    field_params: Dict[str, Dict[str, Any]] = Field(
        default_factory=dict,
        description=(
            "custom Field overrides for the constructor parameters. See "
            ":func:`pydantic.Fields.Field`"
        ),
    )

    returns: Dict[str, Any] = Field(
        default_factory=dict, description="custom returns overrides."
    )

    return_annotation: Optional[Any] = Field(
        None, description="The annotation for the return value."
    )

    annotations: Dict[str, Any] = Field(
        default_factory=dict,
        description="custom annotations for function parameters",
    )

    template: Template = Field(
        Template(name="function.py"),  # type: ignore
        description=(
            "The :class:`demessaging.template.Template` that is used "
            "to render the function for the generated API."
        ),
    )

    reporter_args: Dict[str, BaseReport] = Field(
        default_factory=dict,
        description="Arguments that use the dasf-progress-api",
    )

    json_schema_extra: Dict[str, Any] = Field(
        default_factory=dict,
        description=(
            "Any extra parameter for the JSON schema export for the function"
        ),
    )


@append_parameter_docs
class ClassConfig(BaseConfig):
    """Configuration class for a backend module class."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    name: str = Field(
        "",
        description=(
            "The name of the function. If empty, this will be taken from the "
            "classes ``__name__`` attribute."
        ),
    )

    init_doc: str = Field(
        "",
        description=(
            "The documentation of the function. If empty, this will be taken "
            "from the classes ``__init__`` method."
        ),
    )

    signature: Optional[inspect.Signature] = Field(
        None,
        description=(
            "The calling signature for the function. If empty, this will be "
            "taken from the function itself."
        ),
    )

    methods: List[str] = Field(
        default_factory=list,
        description="methods to use within the backend modules",
    )

    validators: Dict[str, Any] = Field(
        default_factory=dict,
        description="custom validators for the constructor parameters",
    )

    serializers: Dict[str, Any] = Field(
        default_factory=dict,
        description=(
            "A mapping from function argument to serializer that is of "
            "instance :class:`pydantic.functional_serializers.PlainSerializer`"
            " or :class:`pydantic.functional_serializers.WrapSerializer`."
        ),
    )

    field_params: Dict[str, Dict[str, Any]] = Field(
        default_factory=dict,
        description=(
            "custom Field overrides for the constructor parameters. "
            "See :func:`pydantic.Fields.Field`"
        ),
    )

    annotations: Dict[str, Any] = Field(
        default_factory=dict,
        description="custom annotations for constructor parameters",
    )

    template: Template = Field(
        Template(name="class_.py"),  # type: ignore
        description=(
            "The :class:`demessaging.template.Template` that is used "
            "to render the class for the generated API."
        ),
    )

    reporter_args: Dict[str, BaseReport] = Field(
        default_factory=dict,
        description="Arguments that use the dasf-progress-api",
    )

    json_schema_extra: Dict[str, Any] = Field(
        default_factory=dict,
        description=(
            "Any extra parameter for the JSON schema export for the function"
        ),
    )


@append_parameter_docs
class ListenConfig(BaseConfig):
    """A configuration for the listen command"""

    dump_to: Optional[str] = None
    dump_tool: Optional[str] = None
    cmd: Optional[str] = None


@append_parameter_docs
class ModuleConfig(BaseConfig):
    """Configuration class for a backend module."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # it should be Type[BackendFunction], Type[BaseModel], but that's
    #  not well supported by pydantic
    if TYPE_CHECKING:
        members: List[
            Union[
                str,
                Callable,
                Type[object],
                Type[
                    BackendFunction  # pylint: disable=used-before-assignment  # noqa: E501
                ],
                Type[BackendClass],  # pylint: disable=used-before-assignment
            ]
        ]

    messaging_config: Union[PulsarConfig, WebsocketURLConfig] = Field(
        description="Configuration on how to connect to the message broker."
    )

    listen_config: ListenConfig = Field(default_factory=ListenConfig)

    log_config: LoggingConfig = Field(
        LoggingConfig(), description="Configuration for the logging."
    )

    debug: bool = Field(
        False,
        description=(
            "Run the backend module in debug mode (creates more verbose error "
            "messages)."
        ),
    )

    members: List[Union[str, Callable, Type[object], Any]] = Field(  # type: ignore  # noqa: E501
        default_factory=list, description="List of members for this module"  # type: ignore[arg-type]
    )

    imports: str = Field(
        "",
        description="Imports that should be added to the generate API module.",
    )

    template: Template = Field(
        Template(name="module.py"),  # type: ignore
        description=(
            "The :class:`demessaging.template.Template` that is used "
            "to render the module for the generated API."
        ),
    )

    json_schema_extra: Dict[str, Any] = Field(
        default_factory=dict,
        description=(
            "Any extra parameter for the JSON schema export for the function"
        ),
    )

    @property
    def pulsar_config(self) -> Union[PulsarConfig, WebsocketURLConfig]:
        """DEPRECATED! Get the messaging configuration.

        Please use the ``messaging_config`` attribute of this class."""
        warn(
            "The `pulsar_config` property is deprecated. Please use the "
            "`messaging_config` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.messaging_config



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Core module for generating a de-messaging backend module.

This module defines the base classes to serve a general python module as a
backend module in the DASF.

The most important members are:

.. autosummary::

    main
    ~demessaging.backend.module.BackendModule
    ~demessaging.backend.function.BackendFunction
    ~demessaging.backend.class_.BackendClass
"""
from __future__ import annotations

from typing import Type
from warnings import warn

from demessaging.backend import utils  # noqa: F401
from demessaging.backend.class_ import (  # noqa: F401
    BackendClass,
    ClassAPIModel,
)
from demessaging.backend.function import (  # noqa: F401
    BackendFunction,
    FunctionAPIModel,
    ReturnModel,
)
from demessaging.backend.module import (  # noqa: F401
    BackendModule,
    ModuleAPIModel,
)
from demessaging.config import ModuleConfig, PulsarConfig
from demessaging.utils import build_parameter_docs


def main(
    module_name: str = "__main__", *args, **config_kws
) -> Type[BackendModule]:
    """Main function for starting a backend module from the command line."""
    from demessaging.cli import UNKNOWN_TOPIC, get_parser

    default_config: ModuleConfig

    # handle deprecated parameters
    for kw in list(PulsarConfig.model_fields):
        if kw in config_kws:
            val = config_kws.pop(kw)
            warn(
                f"The {kw} parameter hass been moved to the messaging_config. "
                "If should be specified as "
                f"``main(messaging_config=dict({kw}='{val}'))``,"
                f" not ``main({kw}='{val}')``",
                DeprecationWarning,
                stacklevel=2,
            )
            messaging_config = config_kws.setdefault("messaging_config", {})
            if isinstance(messaging_config, dict):
                messaging_config[kw] = val
            else:
                setattr(messaging_config, kw, val)

    if "config" in config_kws:
        default_config = config_kws.pop("config")
        default_config = default_config.copy(update=config_kws)
    else:
        messaging_config = config_kws.setdefault("messaging_config", {})
        if isinstance(messaging_config, dict):
            messaging_config.setdefault("topic", UNKNOWN_TOPIC)
        default_config = ModuleConfig(**config_kws)

    parser = get_parser(module_name, default_config)

    if args:
        ns = parser.parse_args(args)
    else:
        ns = parser.parse_args()

    ns_d = vars(ns)

    command = ns_d.pop("command", None)
    method_kws = {key: ns_d.pop(key) for key in ns_d.pop("command_params", [])}
    method_name = ns_d.pop("method_name", command)
    module_name = ns_d.pop("module_name")

    config = ModuleConfig(**ns_d)

    Model = BackendModule.create_model(module_name, config=config)

    if command:
        method = getattr(Model, method_name)
        result = method(**method_kws)
        if isinstance(result, ReturnModel):
            # a request has been computed
            result = result.model_dump_json()
        print(result)

    return Model


main.__doc__ += build_parameter_docs(ModuleConfig)  # type: ignore



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

import sys

from demessaging import main
from demessaging.cli import UNKNOWN_MODULE


def _main():
    sys.path.insert(0, ".")
    return main(UNKNOWN_MODULE)


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Base module for a websocket connection."""
from __future__ import annotations

import logging
import random
import string
from abc import ABC
from typing import Optional

import websocket
from pydantic import validate_call

from demessaging import config


def get_random_letters(length: int) -> str:
    return "".join(random.choice(string.ascii_letters) for i in range(length))


logger = logging.getLogger(__name__)


class WebsocketConnection(ABC):
    """Base class to connect to a message broker using a websocket."""

    @validate_call
    def __init__(self, pulsar_config: config.BaseMessagingConfig):
        self.pulsar_config = pulsar_config

    def generate_response_topic(self, topic: Optional[str] = None) -> str:
        topic_name = topic or self.pulsar_config.topic or "anonymous"

        return topic_name + "_" + get_random_letters(8)

    def on_message(self, ws_app: websocket.WebSocketApp, msg):
        raise NotImplementedError

    def on_ping(self, ws_app: websocket.WebSocketApp, payload):
        logger.debug("pinged %s", ws_app.url)

    def on_pong(self, ws_app: websocket.WebSocketApp, payload):
        logger.debug("received pong from %s", ws_app.url)

    def on_close(
        self, ws_app: websocket.WebSocketApp, close_status_code, close_msg
    ):
        logger.debug(
            "Websocket connection to %s closed with status code %s, message %s.",
            ws_app.url,
            close_status_code,
            close_msg,
        )

    def create_websocketapp(
        self,
        subscription: Optional[str] = None,
        topic: Optional[str] = None,
        **app_kws,
    ) -> websocket.WebSocketApp:
        topic_name = topic or self.pulsar_config.topic

        topic_url = self.pulsar_config.get_topic_url(topic_name, subscription)

        app_kws.setdefault("on_message", self.on_message)
        app_kws.setdefault("on_ping", self.on_ping)
        app_kws.setdefault("on_pong", self.on_pong)
        app_kws.setdefault("on_close", self.on_close)

        logger.debug("Creating websocket connection for %s", topic_url)

        app = websocket.WebSocketApp(topic_url, **app_kws)

        return app

    def open_socket(
        self,
        subscription: Optional[str] = None,
        topic: Optional[str] = None,
        **connection_kws,
    ) -> websocket.WebSocket:
        topic_name = topic or self.pulsar_config.topic

        topic_url = self.pulsar_config.get_topic_url(topic_name, subscription)

        sock = websocket.create_connection(topic_url, **connection_kws)

        if sock:
            logger.debug("connection to {0} established".format(topic_url))

        return sock



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Transform a python class into a corresponding pydantic model.

The :class:`BackendClass` model in this module generates subclasses based upon
a python class (similarly as the
:class:`~demessaging.backend.function.BackendFunction` does it for functions).
"""
from __future__ import annotations

import inspect
from textwrap import dedent
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import docstring_parser
from pydantic import Field  # pylint: disable=no-name-in-module
from pydantic import BaseModel, create_model
from pydantic.json_schema import JsonSchemaValue

from demessaging.backend import utils
from demessaging.backend.function import (
    BackendFunction,
    BackendFunctionConfig,
    FunctionAPIModel,
    ReturnModel,
)
from demessaging.config import ClassConfig
from demessaging.utils import append_parameter_docs, merge_config


@append_parameter_docs
class BackendClassConfig(ClassConfig):
    """Configuration class for a backend module class."""

    models: Dict[str, Type[BackendFunction]] = Field(
        default_factory=dict,
        description=(
            "Mapping of method name to the function model for the "
            "methods of this class"
        ),
    )

    Class: Type[object] = Field(
        description="The class that corresponds to this config."
    )

    class_name: str = Field(description="Name of the model class")

    @property
    def method_configs(self) -> List[BackendFunctionConfig]:
        """Get a list of the method configs."""
        return [model.backend_config for model in self.models.values()]

    def update_from_cls(self) -> None:
        """Update the config from the corresponding function."""
        Class = self.Class
        if not self.name:
            self.name = Class.__name__ or ""
        if not self.doc:
            self.doc = dedent(inspect.getdoc(Class) or "")
        if not self.init_doc:
            self.init_doc = dedent(inspect.getdoc(Class.__init__) or "")
        if not self.signature:
            self.signature = inspect.signature(Class.__init__)


class ClassAPIModel(BaseModel):
    """A class in the API suitable for RPC via DASF"""

    name: str = Field(
        description="The name of the class that is used as identifier in the RPC."
    )

    rpc_schema: JsonSchemaValue = Field(
        description="The JSON Schema for the constructor of the class."
    )

    methods: List[FunctionAPIModel] = Field(
        description="The list of methods that this class provides."
    )


@append_parameter_docs
class BackendClass(BaseModel):
    """A basis for class models

    Do not directly instantiate from this class, rather use the
    :meth:`create_model` method.
    """

    backend_config: ClassVar[BackendClassConfig]

    @property
    def return_model(self) -> Type[BaseModel]:
        """The return model of the member function."""
        return self.function.return_model

    if TYPE_CHECKING:
        # added properties for subclasses generated by create_model
        function: BackendFunction

    def __call__(self) -> ReturnModel:
        kws = utils.get_kws(self.backend_config.signature, self)

        func_kws = utils.get_kws(
            self.function.backend_config.signature, self.function
        )

        ini: Any = self.backend_config.Class(**kws)  # type: ignore
        func_name = self.function.func_name

        ret = getattr(ini, func_name)(**func_kws)

        # now update the function model and return it
        function: BackendFunction = self.function  # type: ignore
        return function.return_model.model_validate(ret)  # type: ignore[return-value]

    @classmethod
    def get_constructor_fields(
        cls, Class, config: ClassConfig, class_name: Optional[str]
    ) -> Tuple[Dict[str, Any], BackendClassConfig]:
        sig = inspect.signature(Class.__init__)
        docstring = docstring_parser.parse(Class.__doc__)
        init_docstring = docstring_parser.parse(Class.__init__.__doc__)
        docstring.params.extend(init_docstring.params)
        docstring.meta.extend(init_docstring.meta)

        name = Class.__name__
        if not class_name:
            class_name = utils.snake_to_camel("Class", name)

        config = BackendClassConfig(
            Class=Class,
            class_name=class_name,
            **config.model_copy(deep=True).model_dump(),
        )

        fields = utils.get_fields(name, sig, docstring, config)
        fields["class_name"] = fields.pop("func_name")
        return fields, config

    @classmethod
    def create_model(
        cls,
        Class,
        config: Optional[ClassConfig] = None,
        methods: Optional[
            List[Union[Type[BackendFunction], Callable, str]]
        ] = None,
        class_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Type[BackendClass]:
        """Generate a pydantic model from a class.

        Parameters
        ----------
        func: type
            A class
        config: ClassConfig, optional
            The configuration to use. If given, this overrides the
            ``__pulsar_config__`` of the given `Class`
        methods: list of methods, optional
            A list of methods or model classes generated with
            :func:`FunctionModel`. This overrides the methods in `config` or
            the ``__pulsar_config__`` attribute of `Class`
        class_name: str, optional
            The name for the generated subclass of :class:`pydantic.BaseModel`.
            If not given, the name of `Class` is used
        ``**kwargs``
            Any other parameter for the :func:`pydantic.create_model` function

        Returns
        -------
        Subclass of BackendClass
            The newly generated model that represents this class.
        """
        if config is None:
            config = getattr(Class, "__pulsar_config__", ClassConfig())
        config = cast(ClassConfig, config)

        fields, config = cls.get_constructor_fields(Class, config, class_name)
        class_name = cast(str, config.class_name)
        name = Class.__name__

        if "function" in fields:
            raise ValueError(
                f"`function` must not be an init parameter for {name}!"
            )

        if methods:
            pass
        elif config.methods:
            methods = list(config.methods)
        if not methods:
            names_members = inspect.getmembers(
                Class, predicate=inspect.isfunction
            )
            methods = [t[0] for t in names_members if not t[0].startswith("_")]

        if not methods:
            raise ValueError("No methods of the class have been specified!")

        for method in methods:
            if inspect.isclass(method) and issubclass(method, BackendFunction):  # type: ignore  # noqa: E501
                method_name: str = method.backend_config.name  # type: ignore
                FuncModel: Type[BackendFunction] = cast(
                    Type[BackendFunction], method
                )
            elif callable(method):
                method_name = cast(str, method.__name__)
                FuncModel = BackendFunction.create_model(
                    cast(Callable, method),
                    class_name=utils.snake_to_camel(
                        "Meth", class_name, method_name
                    ),
                )
            else:
                method_name = method
                FuncModel = BackendFunction.create_model(
                    getattr(Class, method_name),
                    class_name=utils.snake_to_camel(
                        "Meth", class_name, method_name
                    ),
                )
            if method_name not in config.models:
                config.models[method_name] = FuncModel

        config.methods = list(config.models)

        models = list(config.models.values())
        function_types = models[0]
        for model in models[1:]:
            function_types = Union[function_types, model]  # type: ignore

        fields["function"] = (
            function_types,
            Field(description="The method to call."),
        )
        kwargs.update(fields)

        desc = utils.get_desc(docstring_parser.parse(Class.__doc__))

        Model: Type[BackendClass] = create_model(  # type: ignore
            class_name,
            __validators__=config.validators,
            __module__=Class.__module__,
            __base__=cls,
            **kwargs,  # type: ignore
        )

        Model.backend_config = config

        config.Class = Class
        config.update_from_cls()

        if desc:
            Model.__doc__ = desc
        else:
            Model.__doc__ = ""

        return Model

    @classmethod
    def _get_constructor_model(cls) -> BaseModel:
        """A convenience method to create a pydantic model for __init__."""
        fields, _ = cls.get_constructor_fields(
            cls.backend_config.Class,
            ClassConfig(
                **cls.backend_config.model_dump(
                    exclude={"Class", "models", "class_name"}
                )
            ),
            cls.backend_config.class_name,
        )
        fields["__doc__"] = cls.__doc__
        return create_model(
            cls.backend_config.class_name,
            __module__=cls.backend_config.Class.__module__,
            **fields,
        )

    @classmethod
    def get_api_info(cls) -> ClassAPIModel:
        """Get the API info on the function."""
        constructor_model = cls._get_constructor_model()
        return ClassAPIModel(
            name=cls.backend_config.name,
            rpc_schema=constructor_model.model_json_schema(),
            methods=[
                method_model.get_api_info()
                for method_model in cls.backend_config.models.values()
            ],
        )

    @classmethod
    def model_json_schema(cls, *args, **kwargs) -> Dict[str, Any]:
        ret = super().model_json_schema(*args, **kwargs)
        if cls.backend_config.json_schema_extra:
            ret = merge_config(ret, cls.backend_config.json_schema_extra)
        return ret


try:
    ClassConfig.model_rebuild()
except AttributeError:
    ClassConfig.update_forward_refs()



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Producer of messages submitted for the message broker."""
from __future__ import annotations

import atexit
import base64
import json
import logging
import queue
import threading
import time
from datetime import datetime
from itertools import count
from typing import TYPE_CHECKING, Any, Dict, Optional

from deprogressapi import BaseReport
from websocket import WebSocketApp

from demessaging.messaging.connection import WebsocketConnection
from demessaging.PulsarMessageConstants import (
    MessageType,
    PropertyKeys,
    Status,
)

if TYPE_CHECKING:
    from demessaging.config import BaseMessagingConfig


logger = logging.getLogger(__name__)


class MessageProducer(WebsocketConnection):
    """Producer class to send requests to a registered backend module (topic)"""

    SOCKET_PING_INTERVAL = 60  # 1min
    RECONNECT_TIMEOUT_SLEEP = 5  # seconds

    out_app: WebSocketApp
    in_app: WebSocketApp

    def __init__(
        self, pulsar_config: BaseMessagingConfig, topic: Optional[str] = None
    ):
        super().__init__(pulsar_config)
        self.subscription_name: str = (
            "python-backend-" + datetime.now().isoformat()
        )
        self.context_counter = count()
        self._message_queues: Dict[int, queue.Queue] = {}

        # establish connections
        out_topic = self.pulsar_config.topic
        # topic override if given
        if topic is not None:
            out_topic = topic
        self.response_topic: str = self.generate_response_topic()
        self.out_topic = out_topic
        self.setup_subscription()

    def setup_subscription(self):
        self.out_app = self.create_websocketapp(
            topic=self.out_topic,
            header=self.pulsar_config.header,
            on_message=self.on_out_message,
        )
        # To be thread safe, we generate the response topic here
        self.in_app = self.create_websocketapp(
            subscription=self.subscription_name,
            topic=self.response_topic,
            header=self.pulsar_config.header,
            on_message=self.on_in_message,
        )

    def connect(self):
        self.out_app_thread = threading.Thread(
            target=self.out_app.run_forever,
            kwargs=dict(
                reconnect=self.RECONNECT_TIMEOUT_SLEEP,
                ping_interval=self.SOCKET_PING_INTERVAL,
            ),
            daemon=True,
        )
        self.in_app_thread = threading.Thread(
            target=self.in_app.run_forever,
            kwargs=dict(
                reconnect=self.RECONNECT_TIMEOUT_SLEEP,
                ping_interval=self.SOCKET_PING_INTERVAL,
            ),
            daemon=True,
        )
        self.out_app_thread.start()
        self.in_app_thread.start()
        atexit.register(self.disconnect)

    @property
    def is_connected(self) -> bool:
        """Check if the websocket apps are connected."""
        return bool(
            self.out_app.sock
            and self.out_app.sock.connected
            and self.in_app.sock
            and self.in_app.sock.connected
        )

    def wait_for_connection(self, timeout=10):
        """Wait until the websockets are connected"""
        start = time.time()
        while not self.is_connected and time.time() - start < timeout:
            time.sleep(0.1)
        return self.is_connected

    def on_out_message(self, ws_app: WebSocketApp, ack):
        """Message handler for the outgoing websocket connection."""
        # here we only expect acknowledgement messages.

        ack = json.loads(ack)
        if "context" in ack and "result" in ack:
            if "error" in ack["result"]:
                self._message_queues[int(ack["context"])].put_nowait(
                    {
                        "status": "error",
                        "error": "error sending the request",
                        "msg": ack,
                    }
                )
        else:
            logger.error("Invalid message from outgoing websocket: %s", ack)

    def on_in_message(self, ws_app: WebSocketApp, response):
        """Message handler for the incoming websocket connection."""
        # parse json message
        response = json.loads(response)
        props = response["properties"]

        # acknowledge the response
        self.in_app.send(json.dumps({"messageId": response["messageId"]}))

        # mapping from ids to existing reports
        reports: Dict[str, BaseReport] = {}

        if props[PropertyKeys.MESSAGE_TYPE] == MessageType.PROGRESS:
            # we received a progress report - print and ignore
            # decode progress data
            progress_data = base64.b64decode(response["payload"]).decode(
                "utf-8"
            )

            report = BaseReport.from_payload(progress_data)
            if report.report_id in reports:
                base_report = reports[report.report_id]
                for field in report.model_fields:
                    setattr(base_report, field, getattr(report, field))
            else:
                reports[report.report_id] = base_report = report
            if base_report.status != Status.RUNNING:
                base_report.complete(base_report.status)
            else:
                base_report.submit()
            return

        # assert that we received a 'response' message and check for matching context
        if props[PropertyKeys.MESSAGE_TYPE] != MessageType.RESPONSE:
            msg = {
                "status": "error",
                "error": "received message is not a response to the sent request",
                "msg": response,
            }
        elif "info" in props:
            msg = {
                "status": props.get("status", "success"),
                "msg": props["info"],
            }
        elif "api_info" in props:
            msg = {
                "status": props.get("status", "success"),
                "msg": props["api_info"],
            }
        elif "payload" in response:
            status = "success"
            if "status" in props:
                # we might successfully get a response, but it might contain an error from the backend
                status = props["status"]

            # decode b64 payload msg
            payload: str = response["payload"]

            try:
                payload = base64.b64decode(payload).decode("utf-8")
            except Exception as e:
                status = "error"
                payload = "error decoding payload: {0}".format(e)

            if status == "error":
                msg = {
                    "status": "error",
                    "error": payload,
                    "msg": response,
                }
            else:
                msg = {"status": status, "msg": payload}
        else:
            msg = {
                "status": "error",
                "error": "missing response payload",
                "msg": response,
            }
        context = int(props[PropertyKeys.REQUEST_CONTEXT])
        self._message_queues[context].put_nowait(msg)

    async def send_request(self, request_msg) -> Any:
        """Sends the given request to the backend module bound to the topic provided in the pulsar configuration.
        In order to increase re-usability the destination topic can be overridden with the optional topic argument.

        :param request_msg: dictionary providing a 'property' dictionary, a payload string, or both
        :param topic: overrides the used topic for this request
        :return: received response from the backend module
        """
        if not self.is_connected:
            connected = self.wait_for_connection()
            if not connected:
                raise ValueError(
                    "No websocket connection has been established!"
                )
        # create message context (from counter)
        context = next(self.context_counter)
        request_msg["context"] = context
        if "properties" not in request_msg:
            request_msg["properties"] = {}
        request_msg["properties"][
            PropertyKeys.RESPONSE_TOPIC
        ] = self.response_topic
        request_msg["properties"][PropertyKeys.REQUEST_CONTEXT] = request_msg[
            "context"
        ]
        request_msg["properties"].setdefault(
            PropertyKeys.MESSAGE_TYPE, MessageType.REQUEST
        )

        q = self._message_queues[context] = queue.Queue()

        # send message via outgoing connection to request topic
        self.out_app.send(json.dumps(request_msg))

        # wait for the response. In order to be able to use a keyboard
        # interrupt here, we are using timeouts of 1ms
        logger.info("start waiting for message")
        while True:
            try:
                return q.get(timeout=1)  # Allow check for Ctrl-C every second
            except queue.Empty:
                pass

    def disconnect(self):
        """Disconnect in- and out-sockets."""
        if hasattr(self, "in_app"):
            try:
                self.in_app.close()
            except Exception:
                logger.error(
                    "Failed to close incoming websocket connection for producer",
                    exc_info=True,
                )
            finally:
                del self.in_app

        if hasattr(self, "in_app"):
            try:
                self.out_app.close()
            except Exception:
                logger.error(
                    "Failed to close outgoing websocket connection for producer",
                    exc_info=True,
                )
            finally:
                del self.out_app



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""DASF Methods to connect to the Message Broker."""

from .connection import WebsocketConnection  # noqa: F401
from .consumer import MessageConsumer  # noqa: F401
from .producer import MessageProducer  # noqa: F401



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Transform a python function into a corresponding pydantic model.

The :class:`BackendFunction` model in this module generates subclasses based
upon a python class (similarly as the
:class:`~demessaging.backend.class_.BackendClass` does it for classes).
"""
from __future__ import annotations

import inspect
import warnings
from textwrap import dedent
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Optional,
    Type,
    cast,
)

import docstring_parser
from pydantic import Field  # pylint: disable=no-name-in-module
from pydantic import (
    BaseModel,
    ConfigDict,
    RootModel,
    create_model,
    field_validator,
)
from pydantic.functional_serializers import PlainSerializer
from pydantic.json_schema import JsonSchemaValue
from typing_extensions import Annotated

import demessaging.backend.utils as utils
from demessaging.config import FunctionConfig
from demessaging.utils import append_parameter_docs, merge_config


class ReturnModel(RootModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)


def get_return_model(
    docstring: docstring_parser.Docstring, config: BackendFunctionConfig
) -> Type[BaseModel]:
    """Generate field for the return property.

    Parameters
    ----------
    docstring : docstring_parser.Docstring
        The parser that analyzed the docstring

    Returns
    -------
    Any
        The pydantic field
    """
    return_description = ""
    ret_count: int = 0
    for arg in docstring.meta:
        if (
            isinstance(arg, docstring_parser.DocstringReturns)
            and arg.description
        ):
            return_description += "\n- " + arg.description
            ret_count += 1
    return_description = return_description.strip()
    if ret_count == 1:
        return_description = return_description[2:]

    field_kws: Dict[str, Any] = {"default": None}

    if return_description.strip():
        field_kws["description"] = return_description

    field_kws.update(config.returns)

    ret_field = Field(**field_kws)  # type: ignore

    sig = config.signature

    Model: Type[RootModel]

    create_kws: Dict[str, Any] = {}

    if config.return_annotation is not None:
        annotation = config.return_annotation
    elif sig and sig.return_annotation is not sig.empty:
        if sig.return_annotation is None:
            annotation = Any
        else:
            annotation = sig.return_annotation
    else:
        warnings.warn(
            f"Missing return signature for {config.function.__name__}!",
            RuntimeWarning,
        )
        annotation = Any
    if config.return_serializer is not None:
        serializer = PlainSerializer(
            config.return_serializer, return_type=str, when_used="json"
        )
        if config.return_annotation is None and config.return_validators:
            annotation = Any
        annotation = Annotated[annotation, serializer]

    create_kws["root"] = (annotation, ret_field)
    create_kws["__base__"] = ReturnModel
    if config.return_validators:
        create_kws["__validators__"] = {
            f"root_validator_{i}": field_validator("root")(func)
            for i, func in enumerate(config.return_validators, 1)
        }

    if "description" in field_kws:
        create_kws["__doc__"] = field_kws["description"]

    Model = create_model(
        config.class_name,
        **create_kws,  # type: ignore
    )

    return Model


@append_parameter_docs
class BackendFunctionConfig(FunctionConfig):
    """Configuration class for a backend module function."""

    function: Any = Field(description="The function to call.")

    class_name: str = Field(description="Name of the model class")

    def update_from_function(self) -> None:
        """Update the config from the corresponding function."""
        func = self.function
        if not self.name:
            self.name = func.__name__ or ""
        if not self.doc:
            self.doc = dedent(inspect.getdoc(func) or "")
        if not self.signature:
            self.signature = inspect.signature(func)


class FunctionAPIModel(BaseModel):
    """A class in the API suitable for RPC via DASF"""

    name: str = Field(
        description=(
            "The name of the function that is used as identifier in the RPC."
        )
    )

    rpc_schema: JsonSchemaValue = Field(
        description="The JSON Schema for the function."
    )

    return_schema: JsonSchemaValue = Field(
        description="The JSON Schema for the return value."
    )


@append_parameter_docs
class BackendFunction(BaseModel):
    """A base class for a function model.

    Don't use this model, rather use :meth:`create_model` method to
    generate new models.
    """

    model_config = ConfigDict(
        validate_assignment=True, arbitrary_types_allowed=True
    )

    backend_config: ClassVar[BackendFunctionConfig]

    return_model: ClassVar[Type[BaseModel]]

    if TYPE_CHECKING:
        # added properties for subclasses generated by create_model
        func_name: str

    def __call__(self) -> ReturnModel:  # type: ignore
        kws = utils.get_kws(self.backend_config.signature, self)

        for key in self.backend_config.reporter_args:
            kws[key] = getattr(self, key)

        ret = self.backend_config.function(**kws)

        return self.return_model.model_validate(ret)  # type: ignore[return-value]

    @classmethod
    def create_model(
        cls,
        func: Callable,
        config: Optional[FunctionConfig] = None,
        class_name=None,
        **kwargs,
    ) -> Type[BackendFunction]:
        """Create a new pydantic Model from a function.

        Parameters
        ----------
        func: callable
            A function or method
        config: FunctionConfig, optional
            The configuration to use. If given, this overrides the
            ``__pulsar_config__`` of the given `func`
        class_name: str, optional
            The name for the generated subclass of :class:`pydantic.BaseModel`.
            If not given, the name of `func` is used
        ``**kwargs``
            Any other parameter for the :func:`pydantic.create_model` function

        Returns
        -------
        Subclass of BackendFunction
            The newly generated class that represents this function.
        """
        sig = inspect.signature(func)
        docstring = docstring_parser.parse(func.__doc__)  # type: ignore

        if config is None:
            config = getattr(func, "__pulsar_config__", FunctionConfig())
        config = cast(FunctionConfig, config)

        name = cast(str, func.__name__)
        if not class_name:
            class_name = utils.snake_to_camel("Func", name)

        validators = config.validators.copy()
        config.validators.clear()

        config = BackendFunctionConfig(
            function=func,
            class_name=class_name,
            **config.model_copy(deep=True).model_dump(),
        )
        config.validators = validators

        config.update_from_function()

        fields = utils.get_fields(name, sig, docstring, config)

        desc = utils.get_desc(docstring)

        kwargs.update(fields)

        model_validators: Dict[str, Any] = {}
        for field, functions in config.validators.items():
            for i, func in enumerate(functions, 1):
                model_validators[f"{field}_validator_{i}"] = field_validator(
                    field
                )(func)

        Model: Type[BackendFunction] = create_model(  # type: ignore
            class_name,
            __validators__=model_validators,
            __base__=cls,
            __module__=func.__module__,
            **kwargs,  # type: ignore
        )

        Model.return_model = get_return_model(docstring, config)

        Model.backend_config = config

        if desc:
            Model.__doc__ = desc
        else:
            Model.__doc__ = ""

        return Model

    @classmethod
    def get_api_info(cls) -> FunctionAPIModel:
        """Get the API info on the function."""
        return FunctionAPIModel(
            name=cls.backend_config.name,
            rpc_schema=cls.model_json_schema(),
            return_schema=cls.return_model.model_json_schema(),
        )

    @classmethod
    def model_json_schema(cls, *args, **kwargs) -> Dict[str, Any]:
        ret = super().model_json_schema(*args, **kwargs)
        if cls.backend_config.json_schema_extra:
            ret = merge_config(ret, cls.backend_config.json_schema_extra)
        return ret


try:
    BackendFunctionConfig.model_rebuild()
except AttributeError:
    BackendFunctionConfig.update_forward_refs()



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Command-line options for the backend module."""
from __future__ import annotations

import argparse
import json
import logging
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Type

import yaml

from demessaging.config import (
    LoggingConfig,
    ModuleConfig,
    PulsarConfig,
    WebsocketURLConfig,
)

if TYPE_CHECKING:
    from pydantic import BaseModel

UNKNOWN_TOPIC = "__NOTSET"
UNKNOWN_MODULE = "__NOTSET"


def split_namespace(
    ns: argparse.Namespace,
) -> Tuple[argparse.Namespace, Dict[str, argparse.Namespace]]:
    """Split a namespace into multiple namespaces.

    This utility function takes a namespace and looks for attributes with a
    ``.`` inside. These are then splitted and returned as the second attribute.

    Example
    -------
    Consider the following namespace::

        >>> from argparse import Namespace
        >>> ns = Namespace(**{"main": 1, "sub.main": 2})
        >>> split_namespace(ns)
        (Namespace(main=1), {'sub': Namespace(main=2)})
    """
    main_kws: Dict[str, Any] = {}
    namespace_kws: Dict[str, Dict[str, Any]] = {}
    for key, val in vars(ns).items():
        if "." in key:
            identifier, subkey = key.split(".", 1)
            sub_kws = namespace_kws.setdefault(identifier, {})
            sub_kws[subkey] = val
        else:
            main_kws[key] = val
    namespaces = {
        identifier: argparse.Namespace(**kws)
        for identifier, kws in namespace_kws.items()
    }
    return argparse.Namespace(**main_kws), namespaces


class MessagingArgumentParser(argparse.ArgumentParser):
    """An :class:`argparse.ArgumentParser` for the messaging framework."""

    def parse_known_args(self, args=None, namespace=None):
        ns, remaining_args = super().parse_known_args(args, namespace)
        ns, namespaces = split_namespace(ns)
        if "messaging_config" in namespaces:
            messaging_config = namespaces["messaging_config"]
            wsu_config = namespaces["websocketurl_config"]
            pulsar_config = namespaces["pulsar_config"]
            remaining_messaging_config = (
                wsu_config
                if (
                    wsu_config.websocket_url
                    or (wsu_config.producer_url and wsu_config.consumer_url)
                )
                else pulsar_config
            )
            ns.messaging_config = dict(
                chain(
                    vars(messaging_config).items(),
                    vars(remaining_messaging_config).items(),
                )
            )
        if "log_config" in namespaces:
            ns.log_config = vars(namespaces["log_config"])
        if "listen_config" in namespaces:
            ns.listen_config = vars(namespaces["listen_config"])
        return ns, remaining_args


def get_parser(
    module_name: str = "__main__", config: Optional[ModuleConfig] = None
) -> argparse.ArgumentParser:
    """Generate the command line parser."""

    pulsar_config = PulsarConfig(topic=UNKNOWN_TOPIC)
    websocketurl_config = WebsocketURLConfig(topic=UNKNOWN_TOPIC)

    log_config = LoggingConfig()

    if config is None:
        config = ModuleConfig(
            messaging_config=PulsarConfig(topic=UNKNOWN_TOPIC)
        )
    elif isinstance(config.messaging_config, PulsarConfig):
        pulsar_config = config.messaging_config
    elif isinstance(config.messaging_config, WebsocketURLConfig):
        websocketurl_config = config.messaging_config

    conf_dict = config.model_dump()
    messaging_config = config.messaging_config.model_dump()

    parser = MessagingArgumentParser()

    add = parser.add_argument

    connection_group = parser.add_argument_group(
        "Connection options", "General options for the backend module"
    )

    pulsar_group = parser.add_argument_group(
        "Pulsar connection options",
        "Arguments for connecting to a pulsar. This is the default connection "
        "method unless you specify a websocket-url (see below).",
    )

    websocketurl_group = parser.add_argument_group(
        "Websocket URL group",
        "Arguments for connecting to an arbitrary websocket service.",
    )

    logging_group = parser.add_argument_group(
        "Logging Configuration group",
        "Arguments for configuring the logging within DASF.",
    )

    def desc(name, conf: Type[BaseModel] = ModuleConfig, default: Any = None):
        field_info = conf.model_fields[name]
        desc = field_info.description or ""
        if default or field_info.default:
            desc += " Default: %(default)s"
        return desc

    topic_help = desc("topic", PulsarConfig)
    if messaging_config["topic"] == UNKNOWN_TOPIC:
        topic_help += " This option is required!"
    else:
        topic_help += " Default: %(default)s"

    connection_group.add_argument(
        "-t",
        "--topic",
        help=topic_help,
        required=messaging_config["topic"] == UNKNOWN_TOPIC,
        dest="messaging_config.topic",
        default=messaging_config["topic"],
    )

    connection_group.add_argument(
        "--header",
        help=desc("header", PulsarConfig),
        metavar='{"authorization": "Token ..."}',
        dest="messaging_config.header",
        default=messaging_config["header"],
    )

    module_help = "Name of the backend module."
    if module_name == UNKNOWN_MODULE:
        module_help += " This option is required!"
    else:
        module_help += " Default: %(default)s"

    add(
        "-m",
        "--module",
        default=module_name,
        required=module_name == UNKNOWN_MODULE,
        dest="module_name",
        help=module_help,
    )

    add(
        "-d",
        "--description",
        help=desc("doc"),
        default=conf_dict["doc"],
        dest="doc",
    )

    add(
        "--debug",
        help=desc("debug"),
        action="store_true",
    )

    default_host = messaging_config.get("host", pulsar_config.host)
    pulsar_group.add_argument(
        "-H",
        "--host",
        help=desc("host", PulsarConfig, default_host),
        default=default_host,
        dest="pulsar_config.host",
    )

    default_port = messaging_config.get("port", pulsar_config.port)
    pulsar_group.add_argument(
        "-p",
        "--port",
        help=desc("port", PulsarConfig, default_port),
        default=default_port,
        dest="pulsar_config.port",
    )

    default_persistent = messaging_config.get(
        "persistent", pulsar_config.persistent
    )
    pulsar_group.add_argument(
        "--persistent",
        help=desc("persistent", PulsarConfig, default_persistent),
        default=default_persistent,
        dest="pulsar_config.persistent",
    )

    default_tenant = messaging_config.get("tenant", pulsar_config.tenant)
    pulsar_group.add_argument(
        "--tenant",
        help=desc("tenant", PulsarConfig, default_tenant),
        default=default_tenant,
        dest="pulsar_config.tenant",
    )

    default_namespace = messaging_config.get(
        "namespace", pulsar_config.namespace
    )
    pulsar_group.add_argument(
        "--namespace",
        help=desc("namespace", PulsarConfig, default_namespace),
        default=default_namespace,
        dest="pulsar_config.namespace",
    )

    default_max_workers = messaging_config.get(
        "max_workers", pulsar_config.max_workers
    )
    connection_group.add_argument(
        "--max-workers",
        help=desc("max_workers", PulsarConfig, default_max_workers),
        default=default_max_workers,
        type=int,
        dest="messaging_config.max_workers",
    )

    default_queue_size = messaging_config.get(
        "queue_size", pulsar_config.queue_size
    )
    connection_group.add_argument(
        "--queue_size",
        help=desc("queue_size", PulsarConfig, default_queue_size),
        default=default_queue_size,
        type=int,
        dest="messaging_config.queue_size",
    )

    default_max_payload_size = messaging_config.get(
        "max_payload_size", pulsar_config.max_payload_size
    )
    connection_group.add_argument(
        "--max_payload_size",
        help=desc("max_payload_size", PulsarConfig, default_max_payload_size),
        default=default_max_payload_size,
        type=int,
        dest="messaging_config.max_payload_size",
    )

    default_producer_keep_alive = messaging_config.get(
        "producer_keep_alive", pulsar_config.producer_keep_alive
    )
    connection_group.add_argument(
        "--producer-keep-alive",
        help=desc(
            "producer_keep_alive", PulsarConfig, default_producer_keep_alive
        ),
        default=default_producer_keep_alive,
        type=int,
        dest="messaging_config.producer_keep_alive",
    )

    default_producer_connection_timeout = messaging_config.get(
        "producer_connection_timeout",
        pulsar_config.producer_connection_timeout,
    )
    connection_group.add_argument(
        "--producer-connection-timeout",
        help=desc(
            "producer_connection_timeout",
            PulsarConfig,
            default_producer_connection_timeout,
        ),
        default=default_producer_connection_timeout,
        type=int,
        dest="messaging_config.producer_connection_timeout",
    )

    default_websocket_url = messaging_config.get(
        "websocket_url", websocketurl_config.websocket_url
    )
    websocketurl_group.add_argument(
        "--websocket-url",
        help=desc("websocket_url", WebsocketURLConfig, default_websocket_url),
        default=default_websocket_url,
        dest="websocketurl_config.websocket_url",
    )

    websocketurl_group.add_argument(
        "--producer-url",
        help=desc("producer_url", WebsocketURLConfig, default_websocket_url),
        default=default_websocket_url,
        dest="websocketurl_config.producer_url",
    )

    websocketurl_group.add_argument(
        "--consumer-url",
        help=desc("consumer_url", WebsocketURLConfig, default_websocket_url),
        default=default_websocket_url,
        dest="websocketurl_config.consumer_url",
    )

    add(
        "--members",
        help=desc("members"),
        nargs="+",
        metavar="member",
        default=conf_dict["members"],
    )

    # logging config
    logging_group.add_argument(
        "--log-config",
        default=log_config.config_file,
        dest="log_config.config_file",
        type=Path,
        help=desc("config_file", LoggingConfig),
    )

    logging_group.add_argument(
        "--log-level",
        default=log_config.level,
        dest="log_config.level",
        help=desc("level", LoggingConfig),
        type=int,
        choices=[
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
            logging.CRITICAL,
        ],
    )

    logging_group.add_argument(
        "--log-file",
        default=log_config.logfile,
        type=Path,
        dest="log_config.logfile",
        help=desc("logfile", LoggingConfig),
    )

    logging_group.add_argument(
        "--merge-log-config",
        action="store_true",
        dest="log_config.merge_config",
        help=desc("merge_config", LoggingConfig),
    )

    logging_group.add_argument(
        "--log-overrides",
        dest="log_config.config_overrides",
        type=_load_dict,
        help=(
            "Any valid YAML string, YAML file or JSON file that is merged "
            "into the logging configuration. This option can be used to "
            "quickly override some default logging functionality."
        ),
    )

    # ----------------------------- Subparsers --------------------------------
    subparsers = parser.add_subparsers(dest="command", title="Commands")

    # subparser for testing the connection
    sp = subparsers.add_parser(
        "test-connect",
        help="Connect the backend module to the pulsar message handler.",
    )
    sp.set_defaults(method_name="test_connect")

    # connect subparser (to connect to the pulsar messaging system)
    sp = subparsers.add_parser(
        "listen",
        help="Connect the backend module to the pulsar message handler.",
    )

    sp.add_argument(
        "--dump-to",
        dest="listen_config.dump_to",
        help=(
            "Instead of processing the request, dump it as a file to the given "
            "location. If you need further customization, use ``--dump-tool``."
        ),
    )

    sp.add_argument(
        "--dump-tool",
        dest="listen_config.dump_tool",
        help="""
            Instead of using ``--dump-to``, use this option to run a specific
            command for each request. We will first create a
            temporary file and then run this command as subprocess. This
            parameter requires `--dump-to` and two curly brackets (``{}``) in
            the argument that specify where to insert the target path. Or use
            ``{path}`` or ``{basename}`` or ``{directory}`` for more explicit
            control in your command.

            If you want to process the dumped file further, combine this
            option with ``--cmd``

            **Examples**

            Copy the request to a given location via rsync::

                --dump-tool 'rsync {} .'

            Copy the request via SSH to another server::

                --dump-tool 'scp {} user@machine:/some/folder/'

            Print the request to stdout and delete the temporary file::

                --dump-tool 'cat {path} && rm {path}'
        """,
    )

    sp.add_argument(
        "-c",
        "--cmd",
        dest="listen_config.cmd",
        help="""
            Instead of processing the request here, dump the request as file to
            the disc and run the dedicated command. The specified command must
            contain two curly braces (``{}``) that will be replaced with the
            path or basename of th file. Or use ``{path}``, or ``{basename}``
            or ``{directory}`` for more explicit control in your command.

            **Examples**

            Cat the request (i.e. always return the input to the sender)::

                --cmd 'cat {}'

            Copy the file via scp and run some command to process it on a
            remote machine::

                --dump-tool 'scp {} user@machine:/some/folder/' --cmd 'some-command /some/folder/{basename}'
        """,
    )

    # schema subparser (to print the module schema)
    sp = subparsers.add_parser(
        "schema", help="Print the schema for the backend module."
    )
    sp.add_argument("-i", "--indent", help="Indent the JSON dump.", type=int)
    sp.set_defaults(method_name="schema_json", command_params=["indent"])

    # test subparser (to test the connection to the connect to the pulsar
    # messaging system)
    sp = subparsers.add_parser(
        "send-request", help="Test a request via the messaging system."
    )
    sp.add_argument(
        "request",
        help=(
            "The request to send as YAML string, or the path to a YAML or "
            "JSON file."
        ),
        type=_load_dict,
    )
    sp.set_defaults(method_name="send_request", command_params=["request"])

    # test subparser (to test the connection to the connect to the pulsar
    # messaging system)
    sp = subparsers.add_parser(
        "compute", help="Process a JSON-formatted file as request."
    )
    sp.add_argument(
        "request",
        help=(
            "The request to process as YAML string, or the path to a YAML or "
            "JSON file."
        ),
        type=_load_dict,
    )
    sp.set_defaults(method_name="process_request", command_params=["request"])

    # shell parser. This parser opens a shell to work with the generated Model
    subparsers.add_parser(
        "shell",
        help="Start an IPython shell",
        description=(
            "This command starts an IPython shell where you can access and "
            "work with the generated pydantic Model. This model class is "
            "available via the ``Model`` variable in the shell."
        ),
    )

    # generate parser. This parser renders the backend module and generates a
    # *frontend* API
    sp = subparsers.add_parser(
        "generate",
        help="Generate an API module",
        description=(
            "This command generates an API module that connects to the "
            "backend module via the pulsar and can be used on the client side."
            " We use isort and black to format the generated python file."
        ),
    )

    sp.add_argument(
        "-l",
        "--line-length",
        default=79,
        type=int,
        help=("The line-length for the output API. " "Default: %(default)s"),
    )

    sp.add_argument(
        "--no-formatters",
        action="store_false",
        dest="use_formatters",
        help=(
            "Do not use any formatters (isort, black or autoflake) for the "
            " generated code."
        ),
    )

    sp.add_argument(
        "--no-isort",
        action="store_false",
        dest="use_isort",
        help="Do not use isort for formatting.",
    )

    sp.add_argument(
        "--no-black",
        action="store_false",
        dest="use_black",
        help="Do not use black for formatting.",
    )

    sp.add_argument(
        "--no-autoflake",
        action="store_false",
        dest="use_autoflake",
        help="Do not use autoflake for formatting.",
    )

    sp.set_defaults(
        command_params=[
            "line_length",
            "use_formatters",
            "use_autoflake",
            "use_black",
            "use_isort",
        ]
    )

    return parser


def _load_dict(fname):
    data = yaml.safe_load(fname)
    if isinstance(data, str):  # assume a file and load from disc
        with open(data) as f:
            if data.endswith(".yml") or data.endswith(".yaml"):
                return yaml.safe_load(f)
            elif data.endswith(".json"):
                return json.load(f)
    return data


if __name__ == "__main__":
    parser = get_parser()
    parser.parse_args()



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

import base64
import datetime
import json
import math
import struct
from typing import List

import numpy as np
import numpy.typing as npt
import xarray as xr


def decode_b64_data(data: str) -> List:
    unpack_iter = struct.iter_unpack("<f", base64.b64decode(data))
    return [e[0] for e in unpack_iter]


def decode(netcdfObject: str) -> xr.Dataset:
    """
    decodes the given json string into a xarray dataset.

    :param netcdfObject: str
    :return: xr.Dataset
    """
    # decode json string to dict
    netcdf_json = json.loads(netcdfObject)

    # extract parameters array and extent
    parameters = netcdf_json["parameters"]
    extent = netcdf_json["extent"]

    # init latitude and longitude arrays
    lons: npt.ArrayLike = np.array([], dtype=float)
    lats: npt.ArrayLike = np.array([], dtype=float)

    dataset = {}
    for parameter in parameters:
        dates: List[datetime.datetime] = []
        data = np.array([])
        width = 0
        height = 0

        # iterate over all coverages for this parameter
        for coverage in parameter["coverages"]:
            # decode b64 string into 1d array
            data_array = decode_b64_data(coverage["data"])

            # first time - init the global dimensions
            if width < 1 or height < 1:
                width = coverage["dimension"]["width"]
                height = coverage["dimension"]["height"]

                # respect inverted y axis
                if coverage["dimension"]["inverted"]:
                    lat_start_idx = 1
                    lat_end_idx = 3
                else:
                    lat_start_idx = 3
                    lat_end_idx = 1

                # now that we know the dimensions, we can reconstruct the lon/lat arrays
                lons = np.linspace(
                    extent[0], extent[2], num=width, endpoint=True
                )
                lats = np.linspace(
                    extent[lat_start_idx],
                    extent[lat_end_idx],
                    num=height,
                    endpoint=True,
                )

            # merge single coverage data
            data = np.append(data, data_array)

            # parse and append date
            date = datetime.datetime.strptime(
                coverage["date"], "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            dates.append(date)

        # reshape merged data to 3d spatio-temporal cube
        data = np.reshape(data, (len(dates), height, width))

        # initialize fill parameter
        fill = parameter["fill"]
        if fill is None:
            fill = math.nan

        # create parameter data array
        dataset[parameter["name"]] = xr.DataArray(
            data=data,
            dims=["time", "lat", "lon"],
            coords={"time": dates, "lat": lats, "lon": lons},
            attrs={"_FillValue": fill},
        )

    # create dataset for all parameters, set crs information
    return xr.Dataset(dataset, attrs={"crs": netcdf_json["crs"]})



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0
"""Serializers for xarray datatypes

This module contains serializers for :class:`xarray.Dataset` and
:class:`xarray.DataArray`.
"""
from base64 import b64encode
from typing import Union

import xarray as xr


def encode_xarray(obj: Union[xr.DataArray, xr.Dataset]) -> str:
    """Encode a dataarray or dataset.

    This function uses scipy to encode a :class:`xarray.DataArray` or
    :class:`xarray.Dataset` as b64encoded bytes. To deserialize the object
    again, run::

        from base64 import b64decode

        # for DataArray
        xr.open_dataarray(b64decode(encoded_dataarray.encode("utf-8")))


        # for DataSet
        xr.open_dataset(b64decode(encoded_dataset.encode("utf-8")))

    Parameters
    ----------
    x : Union[xr.DataArray, xr.Dataset]
        The object that should be serialized

    Returns
    -------
    str
        A base64-encoded string decoded as utf-8

    See also
    --------
    demessaging.validators.xarray.validate_dataarray
    demessaging.validators.xarray.validate_dataset
    """
    netcdf_bytes = obj.to_netcdf(format="NETCDF3_CLASSIC")
    return b64encode(netcdf_bytes).decode("utf-8")



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Backend module to transform a python module into a pydantic model.

This module defines the main model in the demessaging framework. It takes a
list of members, or a module, and creates a new Model that can be used to
generate code, connect to the pulsar, and more. See :class:`BackendModule` for
details.
"""
from __future__ import annotations

import atexit
import base64
import datetime as dt
import hashlib
import inspect
import io
import logging
import subprocess as spr
import tempfile
import traceback
from importlib import import_module
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Type,
    Union,
    cast,
)

import docstring_parser
from deprogressapi import BaseReport
from pydantic import Field  # pylint: disable=no-name-in-module
from pydantic import BaseModel, RootModel, ValidationError, create_model
from pydantic.json_schema import JsonSchemaValue

from demessaging.backend import utils
from demessaging.backend.class_ import BackendClass, ClassAPIModel
from demessaging.backend.function import (
    BackendFunction,
    FunctionAPIModel,
    ReturnModel,
)
from demessaging.config import ModuleConfig
from demessaging.messaging.consumer import MessageConsumer
from demessaging.PulsarMessageConstants import PropertyKeys, Status
from demessaging.utils import append_parameter_docs, merge_config

logger = logging.getLogger(__name__)


@append_parameter_docs
class BackendModuleConfig(ModuleConfig):
    """Configuration class for a backend module."""

    # it should be Union[Type[BackendFunction], Type[BackendClass]], but
    # this is not supported by pydantic

    if TYPE_CHECKING:
        models: List[Union[Type[BackendFunction], Type[BackendClass]]]

    models: List[Any] = Field(  # type: ignore
        default_factory=list,  # type: ignore[arg-type]
        description=(
            "a list of function or class models for the members of the "
            "backend module"
        ),
    )

    module: Any = Field(
        description="The imported backend module (or none, if there is none)"
    )

    class_name: str = Field(description="Name of the model class")


class ModuleAPIModel(BaseModel):
    """An model that represants the API of a backend module."""

    classes: List[ClassAPIModel] = Field(
        description="The RPC-enabled classes that this module contains."
    )

    functions: List[FunctionAPIModel] = Field(
        description="The RPC-enabled functions that this module contains."
    )

    rpc_schema: JsonSchemaValue = Field(
        description="The aggregated JSON schema for an RPC call to this module."
    )


ModuleMember = Union[
    Type[BackendFunction], Type[BackendClass], Callable, str, Type[object]
]


@append_parameter_docs
class BackendModule(RootModel):
    """A base class for a backend module.

    Do not directly instantiate from this class, rather use the
    :meth:`create_model` method.
    """

    backend_config: ClassVar[BackendModuleConfig]
    pulsar: ClassVar[MessageConsumer]

    # type that is implemented by subclasses
    root: Union[BackendFunction, BackendClass]

    def __call__(self) -> ReturnModel:
        """Call the selected member of this backend module."""
        return self.root()  # type: ignore

    @classmethod
    def create_model(
        cls,
        module_name: Optional[str] = None,
        members: Optional[List[ModuleMember]] = None,
        config: Optional[ModuleConfig] = None,
        class_name: Optional[str] = None,
        **config_kws,
    ) -> Type[BackendModule]:
        """Generate a module for a backend module.

        Parameters
        ----------
        module_name: str
            The name of the module to import. If none is given, the `members`
            must be specified
        members: list of members
            The list of members that shall be added to this module. It can be
            a list of

            - :class:`~demessaging.backend.function.BackendFunction` classes (
              generated with
              :meth:`~demessaging.backend.function.BackendFunction.create_model`)
            - :class:`~demessaging.backend.class_.BackendClass` classes (
              generated with
              :meth:`~demessaging.backend.class_.BackendClass.create_model`)
            - functions (that will then be transformed using
              :meth:`~demessaging.backend.function.BackendFunction.create_model`)
            - classes (that will then be transformed using
              :meth:`~demessaging.backend.class_.BackendClass.create_model`)
            - strings, in which case they point to the member of the given
              `module_name`
        config: ModuleConfig, optional
            The configuration for the module. If this is not given, you must
            provide ``config_kws`` or define a ``backend_config`` variable
            within the module corresponding to `module_name`.
        class_name: str, optional
            The name for the generated subclass of :class:`pydantic.BaseModel`.
            If not given, the name of `Class` is used
        ``**config_kws``
            An alternative way to specify the configuration for the backend
            module.

        Returns
        -------
        Subclass of BackendFunction
            The newly generated class that represents this module.
        """
        if module_name is not None:
            module: Any = import_module(module_name)
        else:
            module = None

        if members is None and module is None:
            raise ValueError("Either members or module need to be provided!")

        if config and config_kws:
            raise ValueError("Either config or config_kws can be used!")
        if config_kws:
            config = ModuleConfig(**config_kws)
        elif (
            config is None
            and module is not None
            and hasattr(module, "backend_config")
        ):
            config = module.backend_config

        config = cast(ModuleConfig, config)

        # this should not be camelized
        class_name = class_name or module_name or config.messaging_config.topic

        assert config is not None
        config = BackendModuleConfig(
            module=module,
            class_name=class_name,
            **config.model_copy().model_dump(),
        )

        if not members:
            members = list(config.members)
        if not members:
            assert module is not None
            if hasattr(module, "__all__"):
                members = list(module.__all__)
            else:
                functions = inspect.getmembers(
                    module, predicate=inspect.isfunction
                )
                classes = inspect.getmembers(
                    module, predicate=inspect.isfunction
                )
                members = [t[1] for t in functions if not t[0].startswith("_")]
                members += [t[1] for t in classes if not t[0].startswith("_")]

        # finally check if we have any members
        if not members:
            raise ValueError(
                f"Found no members for the given module {module_name}!"
            )
        models: List[Union[Type[BackendFunction], Type[BackendClass]]] = []

        for i, member in enumerate(list(members)):
            member_obj: ModuleMember
            member_model: Union[Type[BackendFunction], Type[BackendClass]]
            if isinstance(member, str):
                member = getattr(module, member)
            if inspect.isclass(member) and issubclass(
                member, (BackendFunction, BackendClass)  # type: ignore
            ):
                member = cast(
                    Union[Type[BackendFunction], Type[BackendClass]], member
                )
                member_model = member
                member_obj = (
                    member.backend_config.Class
                    if issubclass(member, BackendClass)
                    else member.backend_config.function
                )
            elif inspect.isclass(member):
                member_model = BackendClass.create_model(member)
                member_obj = member
            elif callable(member):
                member_model = BackendFunction.create_model(member)
                member_obj = member
            else:
                raise ValueError(
                    f"Cannot transform {member} to a member model!"
                )
            members[i] = member_obj
            models.append(member_model)

        config.members = members
        config.models = models

        if not config.doc and module:
            docstring = docstring_parser.parse(module.__doc__)
            config.doc = utils.get_desc(docstring)

        member_types = models[0]
        for model in models[1:]:
            member_types = Union[member_types, model]  # type: ignore

        kws = {"__module__": module_name} if module_name else {}

        Model: Type[BackendModule] = create_model(  # type: ignore
            class_name,
            __base__=cls,
            root=(member_types, Field(description="The member to call.")),
            **kws,  # type: ignore
        )

        Model.model_config["title"] = config.messaging_config.topic  # type: ignore

        Model.backend_config = config

        # configure logging
        config.log_config.configure_logging()

        if module is not None:
            config.imports += "\n" + utils.get_module_imports(module)

        Model.__doc__ = config.doc

        return Model

    @classmethod
    def test_connect(cls):
        """Connect to the message pulsar."""
        cls.pulsar = consumer = MessageConsumer(
            pulsar_config=cls.backend_config.messaging_config,
            handle_request=cls.handle_message,
            module_info=cls.model_json_schema(),
            api_info=cls.get_api_info(),
        )
        atexit.register(consumer.disconnect)

        consumer.setup_subscription()

    @classmethod
    def get_api_info(cls) -> ModuleAPIModel:
        """Get the API info on the module."""
        return ModuleAPIModel(
            classes=[
                class_.get_api_info()
                for class_ in cls.backend_config.models
                if issubclass(class_, BackendClass)
            ],
            functions=[
                class_.get_api_info()
                for class_ in cls.backend_config.models
                if issubclass(class_, BackendFunction)
            ],
            rpc_schema=cls.model_json_schema(),
        )

    @classmethod
    def listen(cls):
        """_summary_

        Parameters
        ----------
        dump_to : Optional[str], optional
            Instead of processing the request, dump it as a file to the given
            location. If you need further customization, use `--dump-tool`.
        dump_tool : Optional[str], optional
            Instead of using `--dump-to`, use this option to run a specific
            command for each request. We will first create a
            temporary file and then run this command as subprocess. This
            parameter requires `--dump-to` and two curly brackets (``{}``) in
            the argument that specify where to insert the target path. Or use
            ``{path}`` or ``{basename}`` or ``{directory}`` for more explicit
            control in your command.

            If you want to process the dumped file further, combine this
            option with `cmd`
        cmd : Optional[str], optional
            Instead of processing the request here, dump the request as file to
            the disc and run the dedicated command. The specified command must
            contain two curly braces (``{}``) that will be replaced with the
            path or basename of th file. Or use ``{path}``, or ``{basename}``
            or ``{directory}`` for more explicit control in your command.


        Examples
        --------
        Copy the request to a given location via rsync::

            BackendModule.listen(dump_tool='rsync {} .')

        Copy the request via SSH to another server::

            BackendModule.listen(
                dump_tool='scp {} user@machine:/some/folder/'
            )

        Print the request to stdout and delete the temporary file::

            BackendModule.listen(
                dump_tool='cat {path} && rm {path}'
            )

        Cat the request (i.e. always return the input to the sender)::

            BackendModule.listen(cmd='cat {}')

        Copy the file via scp and run some command to process it on a
        remote machine::

            BackendModule.listen(
                dump_tool='scp {} user@machine:/some/folder/',
                cmd='some-command /some/folder/{basename}',
            )
        """
        cls.pulsar = pulsar = MessageConsumer(
            pulsar_config=cls.backend_config.messaging_config,
            handle_request=cls.handle_message,
            module_info=cls.model_json_schema(),
            api_info=cls.get_api_info(),
        )
        atexit.register(pulsar.disconnect)

        pulsar.wait_for_request()

    @classmethod
    def send_request(
        cls: Type[BackendModule],
        request: Union[BackendModule, IO, Dict[str, Any]],
    ) -> BaseModel:
        """Test a request to the backend.

        Parameters
        ----------
        request: dict or file-like object
            A request to the backend module.
        """
        if isinstance(request, io.IOBase):
            model = cls.model_validate_json("\n".join(request.readlines()))  # type: ignore[arg-type]
        elif hasattr(request, "root"):
            request = cast(BackendModule, request)
            model = cls.model_validate(request.root)
        else:
            model = cls.model_validate(request)
        payload = base64.b64encode(
            model.model_dump_json().encode("utf-8")
        ).decode("utf-8")
        request = {
            "properties": {},
            "payload": payload,
        }

        producer = cls.backend_config.messaging_config.producer

        response = utils.run_async(producer.send_request, request)

        status = response[PropertyKeys.STATUS]
        if status == Status.SUCCESS:
            logger.debug("request successful")
            result = response["msg"]
        elif status == Status.ERROR:
            logger.error("request failed: %s", response["msg"])
            raise ValueError(response["error"])
        else:
            raise ValueError("Unknonw status message %s" % (status,))

        return model.root.return_model.model_validate_json(result)

    def compute(self) -> BaseModel:
        """Send this request to the backend module and compute the result.

        This method updates the model inplace.
        """
        response = self.send_request(self)
        return response

    @classmethod
    def process_request(
        cls: Type[BackendModule],
        request: Union[BackendModule, IO, Dict[str, Any]],
    ) -> ReturnModel:
        """Test a request to the backend.

        Parameters
        ----------
        request: dict or file-like object
            The request that we should process.
        """
        if isinstance(request, io.IOBase):
            model = cls.model_validate_json("\n".join(request.readlines()))  # type: ignore[arg-type]
        elif hasattr(request, "root"):
            request = cast(BackendModule, request)
            model = cls.model_validate(request.root)
        else:
            model = cls.model_validate(request)
        return model()

    @classmethod
    def shell(cls):
        """Start a shell with the module defined."""
        from IPython import start_ipython

        start_ipython(argv=[], user_ns=dict(Model=cls))

    @classmethod
    def generate(
        cls,
        line_length: int = 79,
        use_formatters: bool = True,
        use_autoflake: bool = True,
        use_black: bool = True,
        use_isort: bool = True,
    ) -> str:
        """Generate the code for the frontend module."""
        import autoflake
        import black
        import isort

        code = cls.backend_config.render()

        if use_formatters:
            if use_isort:
                code = isort.code(code, float_to_top=True, profile="black")
            if use_black:
                code = black.format_str(
                    code, mode=black.Mode(line_length=line_length)
                )

            # remove unused imports
            if use_autoflake:
                code = autoflake.fix_code(code, remove_all_unused_imports=True)

            if use_isort:
                code = isort.code(code, float_to_top=True, profile="black")

        if cls.backend_config.module:
            # remove __main__, etc.
            name = cls.backend_config.module.__name__
            code = code.replace(name + ".", "")

        return code.strip() + "\n"

    @classmethod
    def handle_message(
        cls,
        request_msg: dict,
    ):
        """_summary_

        Parameters
        ----------
        request_msg : dict
            The message to handle
        """
        message_id = request_msg["messageId"]
        logger.info("processing request %s", message_id)

        def handle_error(header: str, e: Exception):
            logger.error(
                "Failed processing request %s", message_id, exc_info=True
            )
            if cls.backend_config.debug:
                msg = traceback.format_exc()
            else:
                msg = str(e)
            cls.pulsar.send_error(
                request=request_msg,
                error_message="{}: {}".format(header, msg),
            )

        payload = base64.b64decode(request_msg["payload"]).decode("utf-8")

        listen_config = cls.backend_config.listen_config

        try:
            model = cls.model_validate_json(payload)
        except ValidationError as e:
            handle_error("error validating request", e)
        except Exception as e:
            handle_error("error processing request", e)
        else:
            # dump the file if necessary
            hash = hashlib.sha256(payload.encode("utf-8")).hexdigest()
            request_file_name = "%s_%s_%s.json" % (
                utils.slugify(dt.datetime.now().isoformat()),
                utils.slugify(message_id),
                hash,
            )
            if listen_config.dump_to:
                target_file = Path(listen_config.dump_to) / request_file_name
                logger.debug(
                    "Dumping payload of %i to %s", message_id, target_file
                )
                with target_file.open("w") as f:
                    f.write(payload)
            try:
                if listen_config.dump_tool or listen_config.cmd:
                    with tempfile.TemporaryDirectory() as tmpdir:
                        target_file = Path(tmpdir) / request_file_name
                        with target_file.open("w") as f:
                            f.write(payload)
                        if listen_config.dump_tool:
                            logger.debug(
                                "Running dump_tool for %s", target_file
                            )
                            dump_cmd = cls._format_cmd(
                                listen_config.dump_tool, target_file
                            )
                            logger.debug("    --> %s", dump_cmd)
                            spr.check_output(
                                dump_cmd, stderr=spr.STDOUT, shell=True
                            )
                        if listen_config.cmd:
                            compiled = cls._format_cmd(
                                listen_config.cmd, target_file
                            )
                            result = cls._process_cmd(
                                request_msg, model, compiled
                            )
                        else:
                            result = cls._process_model(request_msg, model)
                else:
                    result = cls._process_model(request_msg, model)
            except Exception as e:
                handle_error("error executing request", e)
            else:
                cls.pulsar.send_response(
                    request=request_msg,
                    response_properties={PropertyKeys.STATUS: Status.SUCCESS},
                    response_payload=result.model_dump_json(),
                )

    @classmethod
    def _format_cmd(cls, template: str, target_file: Path) -> str:
        """Utility function to turn a template into a command-line command."""
        return template.format(
            str(target_file),
            path=str(target_file),
            basename=target_file.name,
            directory=target_file.parent,
        )

    @classmethod
    def _process_model(cls, request_msg, model) -> ReturnModel:
        reporter_args = model.root.backend_config.reporter_args
        for key, reporter in reporter_args.items():
            member_reporter = getattr(model.root, key)
            if member_reporter and isinstance(member_reporter, BaseReport):
                member_reporter._pulsar = cls.pulsar
                member_reporter._request = request_msg
        return model()

    @classmethod
    def _process_cmd(cls, request_msg, model, cmd) -> ReturnModel:
        """Process a shell command"""
        result = spr.check_output(cmd, stderr=spr.STDOUT, shell=True)
        return model.root.return_model.model_validate_json(
            result.decode("utf-8")
        )

    @classmethod
    def model_json_schema(cls, *args, **kwargs) -> Dict[str, Any]:
        ret = super().model_json_schema(*args, **kwargs)
        if cls.backend_config.json_schema_extra:
            ret = merge_config(ret, cls.backend_config.json_schema_extra)
        return ret


try:
    ModuleConfig.model_rebuild()
except AttributeError:
    ModuleConfig.update_forward_refs()



# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Test functions for the logging."""

import logging
from pathlib import Path

from demessaging.config import LoggingConfig


def test_default_config():
    """Test the default logging config."""
    config = LoggingConfig()
    config_dict = config.config_dict
    assert isinstance(config_dict, dict)
    assert "version" in config_dict
    assert "formatters" in config_dict
    assert "handlers" in config_dict
    assert "loggers" in config_dict


def test_log_level_overwrite():
    """Test setting the log level"""
    config = LoggingConfig(level=logging.DEBUG)
    config_dict = config.config_dict
    assert config_dict["loggers"]["demessaging"]["level"] == logging.DEBUG


def test_log_file():
    """Test setting the log level"""
    config = LoggingConfig(logfile=Path("./somefile.log"))
    config_dict = config.config_dict
    assert "file_handler" in config_dict["handlers"]
    assert "file_handler" in config_dict["loggers"]["demessaging"]["handlers"]


def test_other_log_config():
    testfile = Path(__file__).parent / "logging_test.yaml"
    config = LoggingConfig(config_file=testfile)
    config_dict = config.config_dict
    assert "root" in config_dict["loggers"]
    assert "demessaging" in config_dict["loggers"]
    assert "console" not in config_dict["handlers"]


def test_log_config_merge():
    """Test merging the logging config."""
    testfile = Path(__file__).parent / "logging_test.yaml"
    config = LoggingConfig(config_file=testfile, merge_config=True)
    config_dict = config.config_dict
    assert "root" in config_dict["loggers"]
    assert "demessaging" in config_dict["loggers"]
    assert "console" in config_dict["handlers"]
    assert "console" in config_dict["loggers"]["demessaging"]["handlers"]
    assert "handler" in config_dict["loggers"]["demessaging"]["handlers"]


def test_config_overrides():
    """Test overriding the logging config."""
    config = LoggingConfig(
        config_overrides={"loggers": {"demessaging": {"level": logging.DEBUG}}}
    )
    config_dict = config.config_dict
    assert config_dict["loggers"]["demessaging"]["level"] == logging.DEBUG



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: CC0-1.0

"""Setup script for the dasf-messaging-python package."""

import versioneer
from setuptools import setup

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Consumer for messages submitted via the message broker."""
from __future__ import annotations

import asyncio
import base64
import concurrent.futures
import json
import logging
import sys
import textwrap
import threading
import time
from datetime import datetime
from time import sleep
from typing import TYPE_CHECKING, Dict, List, Optional

import websocket

from demessaging.messaging.connection import WebsocketConnection
from demessaging.PulsarMessageConstants import MessageType, PropertyKeys

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from demessaging.backend.module import ModuleAPIModel
    from demessaging.config import BaseMessagingConfig

# patch the asyncio loop if we are on windows
# see https://github.com/tornadoweb/tornado/issues/2751
if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


class MessageConsumer(WebsocketConnection):
    """Consumer for messages submitted via the message broker."""

    RECONNECT_TIMEOUT_SLEEP = 10  # seconds
    SOCKET_PING_INTERVAL = 60  # 1min

    def __init__(
        self,
        pulsar_config: BaseMessagingConfig,
        handle_request,
        handle_response=None,
        module_info: Optional[dict] = None,
        api_info: Optional[ModuleAPIModel] = None,
    ):
        super().__init__(pulsar_config)
        pulsar_config = self.pulsar_config
        self.handle_request = handle_request
        self.handle_response = handle_response
        self.module_info = module_info
        self.api_info = api_info

        # init event loop
        self.loop = asyncio.get_event_loop()
        self.request_semaphore: Optional[threading.BoundedSemaphore] = None
        if pulsar_config.queue_size is not None:
            self.request_semaphore = threading.BoundedSemaphore(
                value=pulsar_config.queue_size
            )
        self.pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=pulsar_config.max_workers
        )

        self.max_payload_size = pulsar_config.max_payload_size
        self.connection_attempts = 0
        self.subscription = None
        self.producers: Dict[str, websocket.WebSocketApp] = {}
        self.producer_threads: Dict[
            websocket.WebSocketApp, threading.Thread
        ] = {}
        self.producer_locks: Dict[websocket.WebSocketApp, threading.Lock] = {}
        self.producer_timer: Dict[websocket.WebSocketApp, threading.Timer] = {}
        self.setup_subscription()

    def wait_for_websocket_connection(
        self, ws_app: websocket.WebSocketApp, timeout=10
    ):
        start = time.time()
        while (
            not (ws_app.sock and ws_app.sock.connected)
            and time.time() - start < timeout
        ):
            time.sleep(0.1)
        return ws_app.sock and ws_app.sock.connected

    def open_producer_app(self, topic, **kwargs):
        ws_app = self.create_websocketapp(
            topic=topic, on_message=self.on_producer_message, **kwargs
        )

        self.producer_threads[ws_app] = thread = threading.Thread(
            target=ws_app.run_forever,
            kwargs=dict(
                reconnect=self.RECONNECT_TIMEOUT_SLEEP,
                ping_interval=self.SOCKET_PING_INTERVAL,
            ),
            daemon=True,
        )
        thread.start()
        self.producer_locks[ws_app] = threading.Lock()
        self.reset_close_timer(ws_app)
        return ws_app

    def close_websocket_app(
        self, ws_app: websocket.WebSocketApp, reason: str = "unspecified"
    ):
        logger.info(
            "Closing producer connection to %s. Reason: %s", ws_app.url, reason
        )
        for topic, _ws_app in list(self.producers.items()):
            if _ws_app is ws_app:
                self.producers.pop(topic, None)
        ws_app.close()
        try:
            timer = self.producer_timer.pop(ws_app)
        except KeyError:
            pass
        else:
            timer.cancel()
        self.producer_threads.pop(ws_app, None)
        self.producer_locks.pop(ws_app, None)

    def reset_close_timer(self, ws_app: websocket.WebSocketApp):
        try:
            current_timer = self.producer_timer[ws_app]
        except KeyError:
            pass
        else:
            current_timer.cancel()
        if self.pulsar_config.producer_keep_alive:
            logger.debug(
                "Resetting producer websocket closing timer for %s", ws_app.url
            )
            keep_alive = self.pulsar_config.producer_keep_alive
            self.producer_timer[ws_app] = timer = threading.Timer(
                keep_alive,
                self.close_websocket_app,
                args=(ws_app, "no message since %s seconds" % keep_alive),
            )
            timer.start()

    def on_producer_message(self, ws_app: websocket.WebSocketApp, msg: str):
        # handle message acknowledgement for producer and reset timer
        ack = json.loads(msg)
        if ack["result"] != "ok":
            logger.error("Failed to send message: %s", ack)

        if not self.pulsar_config.producer_keep_alive:
            self.close_websocket_app(ws_app, "Response successfully sent.")

    def setup_subscription(self):
        # prepare timestamp string as part of subscription name
        timestr = datetime.now().isoformat()[:19]
        subscription_name = "backend-module-" + timestr

        self.connection_attempts += 1
        self.subscription = self.create_websocketapp(
            subscription=subscription_name,
            header=self.pulsar_config.header,
        )

    def disconnect(self):
        # close consumer
        if self.subscription is not None:
            try:
                self.subscription.close()
            except Exception:
                logger.error(
                    "Failed to close incoming websocket connection for consumer",
                    exc_info=True,
                )
            finally:
                self.subscription = None

        # close producers
        for ws_app in list(self.producers.values()):
            self.close_websocket_app(ws_app, "Disconnecting.")

    def wait_for_request(self):
        # register request event handler
        logger.info("waiting for incoming request")
        if self.subscription is None:
            self.setup_subscription()

        try:
            teardown = self.subscription.run_forever(
                ping_interval=self.SOCKET_PING_INTERVAL,
                reconnect=self.RECONNECT_TIMEOUT_SLEEP,
            )
        except Exception:
            logger.error("Exception occured in run_forever", exc_info=True)
            teardown = True
        finally:
            logger.info("Connection dropped with error: %s", teardown)
            if self.loop.is_running():
                self.loop.stop()
            self.disconnect()

    def on_message(self, ws_app: websocket.WebSocketApp, msg):
        # handle empty message
        if msg is None:
            logger.debug("empty message received")
            return

        # parse json message
        msg = json.loads(msg)

        # validate message
        # verify that we got a response_topic
        if MessageConsumer.is_valid_request(msg):
            # acknowledge request
            # FIXME: when do we actually acknowledge a message? right after receiving it or after processing it?
            self.acknowledge(msg)

            # handle according to message type
            msg_type = MessageConsumer.extract_message_type(msg)
            if msg_type == MessageType.PING:
                # simply reply with pong
                self.send_pong(msg)
            elif msg_type == MessageType.PONG:
                # handle pong message
                self.handle_pong(msg)
            elif msg_type == MessageType.INFO:
                # handle info message
                self.handle_info(msg)
            elif msg_type == MessageType.API_INFO:
                # handle API info message
                self.handle_api_info(msg)
            elif msg_type == MessageType.REQUEST:
                # handle request message later via event loop
                # self.loop.call_soon(self.handle_request, msg)
                if self.request_semaphore is None:
                    # no bounded queue - directly pass the request to the executor
                    self.loop.run_in_executor(
                        self.pool, self.handle_request, msg
                    )
                else:
                    # bounded queue size - check available space
                    if self.request_semaphore.acquire(blocking=False):
                        # there is still space in the queue - pass to executor queue
                        self.loop.run_in_executor(
                            self.pool, self.handle_request_via_queue, msg
                        )
                    else:
                        # queue is full - reject request
                        self.send_error(
                            request=msg,
                            error_message="request rejected due to queue overflow, try again later",
                        )

                # todo: do we need to address any exceptions here?? e.g. via future.add_done_callback()
                # Process(target=self.handle_request, args=(msg,)).start()
            elif msg_type == MessageType.RESPONSE:
                if self.handle_response:
                    # handle response message later via event loop
                    self.loop.call_soon(self.handle_response, msg)
                    # Process(target=self.handle_response, args=(msg,)).start()
                else:
                    logger.debug(
                        "ignoring response message due to missing handle_response function"
                    )
            else:
                logger.warning(
                    "received unsupported message type: " + msg_type
                )
        else:
            logger.warning(
                "message with unsupported structure received - ignoring it"
            )

    def handle_request_via_queue(self, msg):
        # here the semaphore has already been acquired
        try:
            self.handle_request(msg)
        finally:
            # no matter what - release the semaphore
            self.request_semaphore.release()

    def acknowledge(self, msg):
        self.subscription.sock.send(
            json.dumps({"messageId": msg["messageId"]})
        )

    def send_error(self, request, error_message):
        logger.info(
            "Sending error message to %s: %s",
            MessageConsumer.extract_response_topic(request),
            error_message,
        )
        self.send_response(
            request=request,
            response_payload=error_message,
            response_properties={"status": "error"},
        )

    def send_response(
        self,
        request,
        response_payload=None,
        msg_type=MessageType.RESPONSE,
        response_properties=None,
    ):
        if not MessageConsumer.is_valid_request(request):
            return

        # the request is valid - create a producer for the given response topic
        response_topic = MessageConsumer.extract_response_topic(request)

        if response_properties is None:
            response_properties = {}

        # prepare response
        response_properties[
            PropertyKeys.REQUEST_CONTEXT
        ] = context = MessageConsumer.extract_context(request)
        response_properties[PropertyKeys.MESSAGE_TYPE] = msg_type
        response_properties[
            PropertyKeys.SOURCE_TOPIC
        ] = self.pulsar_config.topic

        if response_topic in self.producers:
            # we already have a producer for this
            producer = self.producers[response_topic]
        else:
            # no producer yet, create one
            producer = self.open_producer_app(
                topic=response_topic, header=self.pulsar_config.header
            )
            self.producers[response_topic] = producer

        with self.producer_locks[producer]:
            is_connected = self.wait_for_websocket_connection(
                producer, self.pulsar_config.producer_connection_timeout
            )
            if not is_connected:
                logger.error(
                    "Failed to establish producer websocket connection to "
                    "topic %s within %s seconds. Stop sending response.",
                    response_topic,
                    self.pulsar_config.producer_connection_timeout,
                )
                self.close_websocket_app(
                    producer, "Failed to establish websocket connection."
                )
                return

            self.reset_close_timer(producer)

            # send the response
            msg = {
                "properties": response_properties,
                "context": str(context),
                "payload": "",
            }
            if response_payload:
                if isinstance(response_payload, str):
                    response_payload = response_payload.encode("utf-8")
                elif isinstance(response_payload, dict):
                    response_payload = json.dumps(response_payload).encode(
                        "utf-8"
                    )

                if len(response_payload) > self.max_payload_size:
                    logger.debug("sending fragmented message")
                    # the payload exceeds the maximum size - send fragmented message
                    payload_fragments: List[str] = textwrap.wrap(  # type: ignore
                        text=base64.b64encode(response_payload).decode(
                            "utf-8"
                        ),
                        width=self.max_payload_size,
                    )

                    num_fragments = len(payload_fragments)
                    for i in range(num_fragments):
                        fragment_props = response_properties.copy()
                        fragment_props[PropertyKeys.FRAGMENT] = i
                        fragment_props[
                            PropertyKeys.NUM_FRAGMENTS
                        ] = num_fragments

                        fragmented_msg = {
                            "properties": fragment_props,
                            "payload": payload_fragments[i],
                            "context": str(context),
                        }
                        producer.send(json.dumps(fragmented_msg))

                        # sleep 100ms
                        sleep(0.05)

                    logger.debug("sent {} fragments".format(num_fragments))
                    return

                # send unfragmented message
                msg["payload"] = base64.b64encode(response_payload).decode(
                    "utf-8"
                )

            producer.send(json.dumps(msg))

    def send_pong(self, request):
        logger.debug(
            "Sending DASF pong to %s",
            MessageConsumer.extract_response_topic(request),
        )
        self.send_response(request=request, msg_type=MessageType.PONG)

    def handle_pong(self, request):
        logger.info("pong received {0}", request)

    def handle_info(self, info_request):
        # check for available module info
        if self.module_info is None:
            # no module info provided
            logger.debug("missing info")
            self.send_response(
                info_request,
                response_properties={
                    "info": "This module does not provide capability information."
                },
            )
            return

        logger.debug("sending info response...")
        self.send_response(
            request=info_request,
            response_properties={"info": json.dumps(self.module_info)},
        )

    def handle_api_info(self, api_info_request):
        """Show the api of the module."""
        if self.api_info is None:
            # no module info provided
            logger.debug("missing api info")
            self.send_response(
                api_info_request,
                response_properties={
                    "info": "This module does not provide api information."
                },
            )
            return

        logger.debug("sending api_info response...")
        self.send_response(
            request=api_info_request,
            response_properties={"api_info": self.api_info.model_dump_json()},
        )

    @staticmethod
    def extract_response_topic(msg):
        if PropertyKeys.RESPONSE_TOPIC in msg["properties"]:
            return msg["properties"][PropertyKeys.RESPONSE_TOPIC]
        else:
            return None

    @staticmethod
    def extract_context(msg) -> Optional[str]:
        if PropertyKeys.REQUEST_CONTEXT in msg["properties"]:
            return str(msg["properties"][PropertyKeys.REQUEST_CONTEXT])
        elif PropertyKeys.REQUEST_CONTEXT in msg:
            return str(msg[PropertyKeys.REQUEST_CONTEXT])
        else:
            return None

    @staticmethod
    def extract_message_type(msg):
        if PropertyKeys.MESSAGE_TYPE in msg["properties"]:
            return msg["properties"][PropertyKeys.MESSAGE_TYPE]
        else:
            return None

    @staticmethod
    def is_valid_value(value):
        return value is not None and isinstance(value, str) and len(value) > 0

    @staticmethod
    def is_valid_request(request_message):
        # for now the request is valid if we find a valid response topic, context and message type
        return (
            MessageConsumer.is_valid_value(
                MessageConsumer.extract_response_topic(request_message)
            )
            and MessageConsumer.is_valid_value(
                MessageConsumer.extract_context(request_message)
            )
            and MessageConsumer.is_valid_value(
                MessageConsumer.extract_message_type(request_message)
            )
        )



"""Configuration module for logging in DASF.

This module defines the configuration class for logging within the DASF
Framework.
"""

# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

import logging.config
from pathlib import Path
from typing import Dict, Optional

import yaml
from pydantic import Field, FilePath, PositiveInt
from pydantic_settings import BaseSettings, SettingsConfigDict

from demessaging.utils import append_parameter_docs, merge_config

DEFAULT_CONFIG_FILE = Path(__file__).parent / "logging.yaml"


@append_parameter_docs
class LoggingConfig(BaseSettings):  # type: ignore
    """Configuration for logging."""

    model_config = SettingsConfigDict(env_prefix="de_logging_")

    config_file: FilePath = Field(
        DEFAULT_CONFIG_FILE,
        description="Path to the logging configuration.",
    )

    level: Optional[PositiveInt] = Field(
        None,
        description=(
            "Level for the logger. Setting this will override any levels "
            "specified in the logging config file. The lower the value, the "
            "more verbose the logging. Typical levels are 10 (DEBUG), "
            "20 (INFO), 30 (WARNING), 40 (ERROR) and 50 (CRITICAL)."
        ),
    )

    logfile: Optional[Path] = Field(
        None,
        description=(
            "A path to use for logging. If this is specified, we will add a "
            "RotatingFileHandler that loggs to the given path and add this "
            "handler to any logger in the logging config."
        ),
    )

    config_overrides: Optional[Dict] = Field(
        None,
        description=(
            "A dictionary to override the configuration specified in the "
            "logging configuration file."
        ),
    )

    merge_config: bool = Field(
        False,
        description=(
            "If this is True, the specified logging configuration file will "
            "be merged with the default one at %s"
        )
        % DEFAULT_CONFIG_FILE,
    )

    @property
    def config_dict(self) -> Dict:
        """Configuration dictionary for the logging."""
        config_file = self.config_file
        with config_file.open("rt") as f:
            config = yaml.safe_load(f.read())
        if not config:
            raise ValueError("Config file at %s is empty!" % (config_file,))

        if self.merge_config and not config_file.samefile(DEFAULT_CONFIG_FILE):
            # merge the specified config into the default config.
            with DEFAULT_CONFIG_FILE.open("rt") as f:
                default_config = yaml.safe_load(f.read())
            config = merge_config(default_config, config)

        if self.level is not None:
            for logger_config in config.get("loggers", {}).values():
                logger_config["level"] = self.level
        if self.logfile is not None:
            formatter = {
                "full": {
                    "format": "%(asctime)s - %(levelname)s - [%(name)s.%(funcName)s]: %(message)s"
                }
            }
            handler_config = {
                "class": "logging.handlers.RotatingFileHandler",
                "mode": "w",
                "level": "DEBUG",
                "formatter": "full",
                "filename": str(self.logfile),
                "maxBytes": 10485760,  # 10MB
                "backupCount": 5,
                "encoding": "utf8",
                "delay": True,
            }
            merge_config(
                config,
                {
                    "formatters": formatter,
                    "handlers": {"file_handler": handler_config},
                },
            )
            for logger_config in config.get("loggers", {}).values():
                logger_config.setdefault("handlers", []).append("file_handler")

        if self.config_overrides:
            config = merge_config(config, self.config_overrides)

        return config

    def configure_logging(self):
        """Configure the loggers based upon the given config."""
        logging.config.dictConfig(self.config_dict)



"""DASF API utilties for to access the configuration."""

# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import inspect
from typing import Any, Callable, Optional, Type, TypeVar, Union

from demessaging.config.registry import ApiRegistry
from demessaging.utils import merge_config

T = TypeVar("T", bound=Callable[..., Any])


#: registry for the stuff that should be available in the generated client stub
registry = ApiRegistry()


def configure(
    js: Optional[str] = None, merge: bool = True, **kwargs
) -> Callable[[T], T]:
    """Configuration decorator for function or modules.

    Use this function as a decorator for classes or functions in the backend
    module like so::

        >>> @configure(field_params={"a": {"gt": 0}}, returns={"gt": 0})
        ... def sqrt(a: float) -> float:
        ...     import math
        ...
        ...     return math.sqrt(a)


    The available parameters for this function vary depending on what you
    are decorating. If you are decorating a class, your parameters must be
    valid for the :class:`ClassConfig`. If you are decorating a function, your
    parameters must be valid for a :class:`FunctionConfig`.

    Parameters
    ----------
    js: Optional[str]
        A JSON-formatted string that can be used to setup the config.
    merge: bool
        If True (default), then the configuration will be merged with the
        existing configuration for the function (if existing)
    ``**kwargs``
        Any keyword argument that can be used to setup the config.

    Notes
    -----
    If you are specifying any ``kwargs``, your first argument (`js`) should
    be ``None``.
    """
    from demessaging.config.backend import ClassConfig, FunctionConfig

    def decorator(obj: T) -> T:
        ConfClass: Union[Type[ClassConfig], Type[FunctionConfig]]
        if inspect.isclass(obj):
            ConfClass = ClassConfig
        else:
            ConfClass = FunctionConfig
        if js and kwargs:
            raise ValueError(
                "You can either specify a JSON string or keyword arguments, "
                "not both!"
            )
        if js:
            config = ConfClass.model_validate_json(js)
        else:
            config = ConfClass(**kwargs)
        if merge and hasattr(obj, "__pulsar_config__"):
            old = obj.__pulsar_config__.model_dump()
            new = config.model_dump()
            config = ConfClass(**merge_config(old, new))

        obj.__pulsar_config__ = config  # type: ignore
        return obj

    return decorator



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

from warnings import warn

from demessaging.messaging.constants import *  # noqa: F403, F401

warn(
    "The demessaging.PulsarMessageConstants module has been renamed to "
    "demessaging.messaging.constants and will be removed soon!",
    DeprecationWarning,
)



# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

from warnings import warn

from demessaging.messaging.producer import MessageProducer  # noqa: F403, F401

warn(
    "The demessaging.PulsarMessageProducer module has been renamed to "
    "demessaging.messaging.producer and will be removed soon!",
    DeprecationWarning,
)


class PulsarMessageProducer(MessageProducer):
    # deprecated

    def __init__(self, *args, **kwargs):
        warn(
            "The `demessaging.PulsarMessageProducer.PulsarMessageProducer` "
            "class has been replaced by the "
            "`demessaging.messaging.producer.MessageProducer` class "
            "and will be removed soon!",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)



"""API registry class for demessaging."""

# SPDX-FileCopyrightText: 2019-2025 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences
# SPDX-FileCopyrightText: 2020-2021 Helmholtz-Zentrum Geesthacht GmbH
# SPDX-FileCopyrightText: 2021-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import importlib
import inspect
from typing import Any, Callable, Dict, List, Optional, TypeVar

from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import Field, field_validator

from demessaging.utils import append_parameter_docs

T = TypeVar("T", bound=Callable[..., Any])


@append_parameter_docs
class ApiRegistry(BaseModel):
    """A registry for imports and encoders"""

    @field_validator("imports")
    @classmethod
    def can_import_import(cls, imports: Dict[str, str]) -> Dict[str, str]:
        errors: List[ImportError] = []
        for key in imports:
            try:
                importlib.import_module(key)
            except ImportError as e:
                errors.append(e)
            except Exception:
                raise
        if errors:
            raise ValueError(
                "Could not import all modules!\n    "
                + "\n    ".join(map(str, errors))
            )
        return imports

    imports: Dict[str, str] = Field(
        default_factory=dict,
        description=(
            "Modules to import at the top of every file. The first "
            "item is the module, the second is the alias"
        ),
    )

    objects: List[str] = Field(
        default_factory=list,
        description=(
            "Source code for objects that should be inlined in the generated "
            "Python API."
        ),
    )

    def register_import(
        self, module: str, alias: Optional[str] = None
    ) -> None:
        """Register a module that needs to be imported in generated API files.

        Parameters
        ----------
        module: str
            The name of the module, e.g. matplotlib.pyplot
        """
        self.imports[module] = alias or ""

    def register_type(self, obj: T) -> T:
        """Register a class or function to be available in the generated API.

        Use this function if you want to have certain functions of classes
        available in the generated API module, but they should not be
        transformed to a call to the backend module.
        """
        self.objects.append(inspect.getsource(obj))
        return obj

    def hard_code(self, python_code: str) -> None:
        """Register some code to be implemented in the generated module.

        Parameters
        ----------
        python_code: str
            The code that is supposed to be executed on a module level.
        """
        self.objects.append(python_code)
        return
